@@ -1,15 +1,15 @@
-// Copyright 2020-2021 Kevin Reid under the terms of the MIT License as detailed
+// Copyright 2020-2022 Kevin Reid under the terms of the MIT License as detailed
 // in the accompanying file README.md or <https://opensource.org/licenses/MIT>.
 
 use cgmath::{EuclideanSpace as _, Point2, Vector2, Vector3, Zero as _};
 use std::collections::{HashMap, HashSet};
 use std::time::Duration;
 
-use crate::apps::Tick;
-use crate::camera::Viewport;
+use crate::camera::{FogOption, GraphicsOptions, LightingOption, TransparencyOption, Viewport};
 use crate::character::Character;
 use crate::listen::{ListenableCell, ListenableSource};
 use crate::math::FreeCoordinate;
+use crate::time::Tick;
 use crate::universe::{URef, Universe};
 
 /// Parse input events, particularly key-down/up pairs, into character control and such.
@@ -86,8 +86,11 @@ impl InputProcessor {
             Key::Down => true,
             Key::Character(' ') => true,
             Key::Character(d) if d.is_ascii_digit() => true,
+            Key::Character('i') => true,
             Key::Character('l') => true,
+            Key::Character('o') => true,
             Key::Character('p') => true,
+            Key::Character('u') => true,
             _ => false,
         }
     }
@@ -97,8 +100,11 @@ impl InputProcessor {
         #[allow(clippy::match_like_matches_macro)]
         match key {
             Key::Character(d) if d.is_ascii_digit() => true,
+            Key::Character('i') => true,
             Key::Character('l') => true,
+            Key::Character('o') => true,
             Key::Character('p') => true,
+            Key::Character('u') => true,
             // TODO: move slot selection commands here
             _ => false,
         }
@@ -258,6 +264,7 @@ impl InputProcessor {
             universe: _,
             character: character_opt,
             paused: paused_opt,
+            graphics_options,
         } = targets;
 
         let dt = tick.delta_t.as_secs_f64();
@@ -288,7 +295,19 @@ impl InputProcessor {
 
         for key in self.command_buffer.drain(..) {
             match key {
+                Key::Character('i') => {
+                    if let Some(cell) = graphics_options {
+                        cell.update_mut(|options| {
+                            options.lighting_display = match options.lighting_display {
+                                LightingOption::None => LightingOption::Flat,
+                                LightingOption::Flat => LightingOption::Smooth,
+                                LightingOption::Smooth => LightingOption::None,
+                            };
+                        });
+                    }
+                }
                 Key::Character('l') => {
+                    // TODO: duplicated with fn toggle_mouselook_mode() because of borrow conflicts
                     let new_state = !*self.mouselook_mode.get();
                     self.mouselook_mode.set(new_state);
                     if new_state {
@@ -296,10 +315,35 @@ impl InputProcessor {
                         self.mouse_previous_pixel_position = None;
                     }
                 }
+                Key::Character('o') => {
+                    if let Some(cell) = graphics_options {
+                        cell.update_mut(|options| {
+                            options.transparency = match options.transparency {
+                                TransparencyOption::Surface => TransparencyOption::Volumetric,
+                                TransparencyOption::Volumetric => {
+                                    TransparencyOption::Threshold(notnan!(0.5))
+                                }
+                                TransparencyOption::Threshold(_) => TransparencyOption::Surface,
+                            };
+                        });
+                    }
+                }
                 Key::Character('p') => {
                     // TODO: bind escape key, focus loss, etc to pause
                     if let Some(paused) = paused_opt {
-                        paused.set(!*paused.get());
+                        paused.update_mut(|p| *p = !*p);
+                    }
+                }
+                Key::Character('u') => {
+                    if let Some(cell) = graphics_options {
+                        cell.update_mut(|options| {
+                            options.fog = match options.fog {
+                                FogOption::None => FogOption::Abrupt,
+                                FogOption::Abrupt => FogOption::Compromise,
+                                FogOption::Compromise => FogOption::Physical,
+                                FogOption::Physical => FogOption::None,
+                            };
+                        });
                     }
                 }
                 Key::Character(numeral) if numeral.is_digit(10) => {
@@ -320,6 +364,16 @@ impl InputProcessor {
         self.mouselook_mode.as_source()
     }
 
+    // TODO: duplicated with the keybinding impl because of borrow conflicts
+    pub(crate) fn toggle_mouselook_mode(&mut self) {
+        let new_state = !*self.mouselook_mode.get();
+        self.mouselook_mode.set(new_state);
+        if new_state {
+            // Clear delta tracking just in case
+            self.mouse_previous_pixel_position = None;
+        }
+    }
+
     /// Returns the position which should be used for click/cursor raycasting.
     /// This is not necessarily equal to the tracked mouse position.
     ///
@@ -356,6 +410,7 @@ pub struct InputTargets<'a> {
     pub universe: Option<&'a mut Universe>,
     pub character: Option<&'a URef<Character>>,
     pub paused: Option<&'a ListenableCell<bool>>,
+    pub graphics_options: Option<&'a ListenableCell<GraphicsOptions>>,
 }
 
 /// A platform-neutral representation of keyboard keys for [`InputProcessor`].
@@ -390,6 +445,7 @@ mod tests {
                 universe: Some(universe),
                 character: Some(character),
                 paused: None,
+                graphics_options: None,
             },
             Tick::arbitrary(),
         );
