@@ -1,27 +1,29 @@
-// Copyright 2020-2021 Kevin Reid under the terms of the MIT License as detailed
+// Copyright 2020-2022 Kevin Reid under the terms of the MIT License as detailed
 // in the accompanying file README.md or <https://opensource.org/licenses/MIT>.
 
 //! Definition of blocks, which are game objects which live in the grid of a
 //! [`Space`]. See [`Block`] for details.
 
 use std::borrow::Cow;
-use std::convert::TryFrom as _;
 use std::fmt;
-use std::ops::Deref;
-use std::sync::Arc;
 
 use cgmath::{EuclideanSpace as _, Point3, Vector4, Zero as _};
 
-use crate::listen::{Gate, Listener, Notifier};
-use crate::math::{FreeCoordinate, GridCoordinate, GridPoint, GridRotation, Rgb, Rgba};
+use crate::listen::Listener;
+use crate::math::{
+    FreeCoordinate, GridCoordinate, GridPoint, GridRotation, OpacityCategory, Rgb, Rgba,
+};
 use crate::raycast::{Ray, Raycaster};
 use crate::space::{Grid, GridArray, SetCubeError, Space, SpaceChange};
-use crate::transaction::{
-    Merge, PreconditionFailed, Transaction, TransactionConflict, Transactional,
-};
 use crate::universe::{RefError, URef};
 use crate::util::{ConciseDebug, CustomFormat};
 
+mod attributes;
+pub use attributes::*;
+
+mod block_def;
+pub use block_def::*;
+
 pub mod builder;
 #[doc(inline)]
 pub use builder::BlockBuilder;
@@ -163,10 +165,7 @@ impl Block {
     #[inline]
     fn evaluate_impl(&self, depth: u8) -> Result<EvaluatedBlock, EvalBlockError> {
         match self {
-            Block::Indirect(def_ref) => def_ref
-                .try_borrow()?
-                .block
-                .evaluate_impl(next_depth(depth)?),
+            Block::Indirect(def_ref) => def_ref.try_borrow()?.evaluate_impl(next_depth(depth)?),
 
             &Block::Atom(ref attributes, color) => Ok(EvaluatedBlock {
                 attributes: attributes.clone(),
@@ -175,6 +174,14 @@ impl Block {
                 resolution: 1,
                 opaque: color.fully_opaque(),
                 visible: !color.fully_transparent(),
+                voxel_opacity_mask: if color.fully_transparent() {
+                    None
+                } else {
+                    Some(
+                        GridArray::from_elements(Grid::for_block(1), [color.opacity_category()])
+                            .unwrap(),
+                    )
+                },
             }),
 
             &Block::Recur {
@@ -195,6 +202,7 @@ impl Block {
                         resolution: 1,
                         opaque: false,
                         visible: false,
+                        voxel_opacity_mask: None,
                     });
                 }
 
@@ -205,67 +213,57 @@ impl Block {
                     .intersection(block_space.grid())
                     .unwrap_or_else(|| Grid::new(offset, [1, 1, 1]) /* arbitrary value */);
 
-                // TODO: The color sum actually needs to be weighted by alpha. (Too bad we're not using premultiplied alpha.)
-                // TODO: Should not be counting interior voxels for the color, only visible surfaces.
-
-                let mut color_sum: Vector4<f32> = Vector4::zero();
                 let voxels = block_space
                     .extract(
                         occupied_grid,
                         #[inline(always)]
                         |_index, sub_block_data, _lighting| {
-                            let sub_evaluated = sub_block_data.evaluated();
-                            color_sum += sub_evaluated.color.into();
-                            Evoxel {
-                                color: sub_evaluated.color,
-                                selectable: sub_evaluated.attributes.selectable,
-                                collision: sub_evaluated.attributes.collision,
-                            }
+                            Evoxel::from_block(sub_block_data.evaluated())
                         },
                     )
                     .translate(-offset.to_vec());
 
-                Ok(EvaluatedBlock {
-                    attributes: attributes.clone(),
-                    // The single color is the mean of the actual block colors.
-                    color: Rgba::try_from(
-                        (color_sum.truncate() / (occupied_grid.volume() as f32))
-                            .extend(color_sum.w / (full_resolution_grid.volume() as f32)),
-                    )
-                    .expect("Recursive block color computation produced NaN"),
+                Ok(EvaluatedBlock::from_voxels(
+                    attributes.clone(),
                     resolution,
-                    // TODO wrong test: we want to see if the _faces_ are all opaque but allow hollows
-                    opaque: occupied_grid == full_resolution_grid
-                        && voxels.grid().interior_iter().all(
-                            #[inline(always)]
-                            |p| voxels[p].color.fully_opaque(),
-                        ),
-                    visible: voxels.grid().interior_iter().any(
-                        #[inline(always)]
-                        |p| !voxels[p].color.fully_transparent(),
-                    ),
-
-                    voxels: Some(voxels),
-                })
+                    voxels,
+                ))
             }
 
             // TODO: this has no unit tests
             Block::Rotated(rotation, block) => {
                 let base = block.evaluate()?;
+                if base.voxels.is_none() && base.voxel_opacity_mask.is_none() {
+                    // Skip computation of transforms
+                    return Ok(base);
+                }
+
+                // TODO: Add a shuffle-in-place rotation operation to GridArray and try implementing this using that, which should have less arithmetic involved than these matrix ops
                 let resolution = base.resolution;
+                let inner_to_outer = rotation.to_positive_octant_matrix(resolution.into());
+                let outer_to_inner = rotation
+                    .inverse()
+                    .to_positive_octant_matrix(resolution.into());
+
                 Ok(EvaluatedBlock {
                     voxels: base.voxels.map(|voxels| {
-                        let inner_to_outer = rotation.to_positive_octant_matrix(resolution.into());
-                        let outer_to_inner = rotation
-                            .inverse()
-                            .to_positive_octant_matrix(resolution.into());
-                        // TODO: Add a shuffle-in-place rotation operation to GridArray and try implementing this using that, which should have less unnecessary arithmetic
                         GridArray::from_fn(
                             voxels.grid().transform(inner_to_outer).unwrap(),
                             |cube| voxels[outer_to_inner.transform_cube(cube)],
                         )
                     }),
-                    ..base
+                    voxel_opacity_mask: base.voxel_opacity_mask.map(|mask| {
+                        GridArray::from_fn(mask.grid().transform(inner_to_outer).unwrap(), |cube| {
+                            mask[outer_to_inner.transform_cube(cube)]
+                        })
+                    }),
+
+                    // Unaffected
+                    attributes: base.attributes,
+                    color: base.color,
+                    resolution,
+                    opaque: base.opaque,
+                    visible: base.visible,
                 })
             }
         }
@@ -394,201 +392,6 @@ impl From<Rgba> for Cow<'_, Block> {
     }
 }
 
-/// Collection of miscellaneous attribute data for blocks that doesn't come in variants.
-///
-/// `BlockAttributes::default()` will produce a reasonable set of defaults for “ordinary”
-/// blocks.
-#[derive(Clone, Eq, Hash, PartialEq)]
-// #[non_exhaustive] // TODO: Make this non_exhaustive but give users a way to construct it easily, possibly via BlockBuilder.
-pub struct BlockAttributes {
-    /// The name that should be displayed to players.
-    ///
-    /// The default value is the empty string. The empty string should be considered a
-    /// reasonable choice for solid-color blocks with no special features.
-    pub display_name: Cow<'static, str>,
-
-    /// Whether players' cursors target it or pass through it.
-    ///
-    /// The default value is `true`.
-    pub selectable: bool,
-
-    /// The effect on a [`Body`](crate::physics::Body) of colliding with this block.
-    ///
-    /// The default value is [`BlockCollision::Hard`].
-    pub collision: BlockCollision,
-
-    /// Light emitted by the block.
-    ///
-    /// The default value is [`Rgb::ZERO`].
-    pub light_emission: Rgb,
-    // TODO: add 'behavior' functionality, if we don't come up with something else
-    pub animation_hint: AnimationHint,
-    // Reminder: When adding new fields, add them to the Debug implementation.
-}
-
-impl fmt::Debug for BlockAttributes {
-    /// Only attributes which differ from the default are shown.
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        if self == &Self::default() {
-            // Avoid the braceless formatting used for structs with literally no fields.
-            write!(f, "BlockAttributes {{}}")
-        } else {
-            let mut s = f.debug_struct("BlockAttributes");
-            if self.display_name != Self::default().display_name {
-                // Unwrap the `Cow` for tidier formatting.
-                s.field("display_name", &&*self.display_name);
-            }
-            if self.selectable != Self::default().selectable {
-                s.field("selectable", &self.selectable);
-            }
-            if self.collision != Self::default().collision {
-                s.field("collision", &self.collision);
-            }
-            if self.light_emission != Self::default().light_emission {
-                s.field("light_emission", &self.light_emission);
-            }
-            if self.animation_hint != Self::default().animation_hint {
-                s.field("animation_hint", &self.animation_hint);
-            }
-            s.finish()
-        }
-    }
-}
-
-impl BlockAttributes {
-    /// Block attributes suitable as default values for in-game use.
-    ///
-    /// This function differs from the [`Default::default`] trait implementation only
-    /// in that it is a `const fn`.
-    pub const fn default() -> BlockAttributes {
-        BlockAttributes {
-            display_name: Cow::Borrowed(""),
-            selectable: true,
-            collision: BlockCollision::Hard,
-            light_emission: Rgb::ZERO,
-            animation_hint: AnimationHint::UNCHANGING,
-        }
-    }
-}
-
-impl Default for BlockAttributes {
-    /// Block attributes suitable as default values for in-game use.
-    fn default() -> BlockAttributes {
-        // Delegate to the inherent impl `const fn`.
-        BlockAttributes::default()
-    }
-}
-
-#[cfg(feature = "arbitrary")]
-impl<'a> arbitrary::Arbitrary<'a> for BlockAttributes {
-    fn arbitrary(u: &mut arbitrary::Unstructured<'a>) -> arbitrary::Result<Self> {
-        Ok(BlockAttributes {
-            display_name: Cow::Owned(u.arbitrary()?),
-            selectable: u.arbitrary()?,
-            collision: u.arbitrary()?,
-            light_emission: u.arbitrary()?,
-            animation_hint: u.arbitrary()?,
-        })
-    }
-
-    fn size_hint(depth: usize) -> (usize, Option<usize>) {
-        arbitrary::size_hint::and_all(&[
-            String::size_hint(depth),
-            bool::size_hint(depth),
-            BlockCollision::size_hint(depth),
-            Rgb::size_hint(depth),
-            AnimationHint::size_hint(depth),
-        ])
-    }
-}
-
-/// Specifies the effect on a [`Body`](crate::physics::Body) of colliding with the
-/// [`Block`] this applies to.
-#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
-#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
-#[non_exhaustive]
-pub enum BlockCollision {
-    /// The block can be passed through; it is not an obstacle (though intersecting it
-    /// might cause other effects not directly part of collision response).
-    None,
-    /// The block is a perfectly solid obstacle occupying its entire bounding cube.
-    ///
-    /// This is the default value used for most blocks. (Caveat: The default might be
-    /// changed to `Recur` if that proves more ergonomic.)
-    Hard,
-    /// Collide with the block's component voxels individually.
-    ///
-    /// If the block does not have voxels then this is equivalent to [`Hard`](Self::Hard).
-    /// (TODO: Voxel collision not yet finished.)
-    Recur,
-    // Future values might include bouncy solid, water-like resistance, force fields, etc.
-}
-
-/// Specifies how the appearance of a [`Block`] might change, for the benefit of rendering
-/// algorithms. This hint applies both to a block's definition changing and to it being
-/// replaced with some successor block.
-#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
-#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
-pub struct AnimationHint {
-    /// Expect that the block might soon be replaced with an unrelated block.
-    /// Suggestion: avoid combining it with other block meshes.
-    pub(crate) expect_replace: bool,
-
-    /// Expect that the block's shape will change; some of its voxels will move between
-    /// the categories “fully opaque”, “fully transparent”, and “in between”.
-    /// Suggestion: use a rendering strategy which is shape-independent.
-    pub(crate) expect_shape_update: bool,
-
-    /// Expect that the block's voxels' colors (and alpha other than the special 0 and 1
-    /// cases) will change.
-    /// Suggestion: prepare to update texturing without unnecesarily regenerating the mesh.
-    pub(crate) expect_color_update: bool,
-}
-
-impl AnimationHint {
-    /// There are no expectations that the block is soon going to change.
-    ///
-    /// This is the default value of this type and within [`BlockAttributes`].
-    pub const UNCHANGING: Self = Self {
-        expect_replace: false,
-        expect_shape_update: false,
-        expect_color_update: false,
-    };
-
-    /// The block is not going to exist in its current form for long.
-    ///
-    /// This suggests using a rendering technique which is comparatively expensive
-    /// per-block but allows it (and any successors that are also `TEMPORARY`) to be added
-    /// and removed cheaply.
-    pub const TEMPORARY: Self = Self {
-        expect_replace: true,
-        ..Self::UNCHANGING
-    };
-
-    /// The block's appearance is expected to change very frequently, but not by replacing
-    /// the block in its [`Space`].
-    ///
-    /// This suggests using a rendering technique which optimizes for not needing to e.g.
-    /// rebuild chunk meshes.
-    pub const CONTINUOUS: Self = Self {
-        expect_color_update: true,
-        expect_shape_update: true,
-        ..Self::UNCHANGING
-    };
-
-    /// Returns whether this block's value for [`EvaluatedBlock::visible`] is likely to
-    /// change from `false` to `true`.
-    fn might_become_visible(&self) -> bool {
-        self.expect_shape_update
-    }
-}
-
-impl Default for AnimationHint {
-    fn default() -> Self {
-        Self::UNCHANGING
-    }
-}
-
 /// Generic 'empty'/'null' block. It is used by [`Space`] to respond to out-of-bounds requests.
 ///
 /// See also [`AIR_EVALUATED`].
@@ -610,12 +413,14 @@ pub const AIR_EVALUATED: EvaluatedBlock = EvaluatedBlock {
     resolution: 1,
     opaque: false,
     visible: false,
+    voxel_opacity_mask: None,
 };
 
 const AIR_ATTRIBUTES: BlockAttributes = BlockAttributes {
     display_name: Cow::Borrowed("<air>"),
     selectable: false,
     collision: BlockCollision::None,
+    rotation_rule: RotationPlacementRule::Never,
     light_emission: Rgb::ZERO,
     animation_hint: AnimationHint::UNCHANGING,
 };
@@ -653,6 +458,15 @@ pub struct EvaluatedBlock {
     /// Whether the block has any voxels/color at all that make it visible; that is, this
     /// is false if the block is completely transparent.
     pub visible: bool,
+    /// The opacity of all voxels. This is redundant with the data  [`Self::voxels`],
+    /// and is provided as a pre-computed convenience that can be cheaply compared with
+    /// other values of the same type.
+    ///
+    /// May be [`None`] if the block is fully invisible. (TODO: This is a kludge to avoid
+    /// obligating [`AIR_EVALUATED`] to allocate at compile time, which is impossible.
+    /// It doesn't harm normal operation because the point of having this is to compare
+    /// block shapes, which is trivial if the block is invisible.)
+    pub(crate) voxel_opacity_mask: Option<GridArray<OpacityCategory>>,
 }
 
 // TODO: Wait, this isn't really what ConciseDebug is for... shouldn't this be a regular impl Debug?
@@ -665,11 +479,55 @@ impl CustomFormat<ConciseDebug> for EvaluatedBlock {
             .field("visible", &self.visible)
             .field("resolution", &self.resolution)
             .field("voxels", &"...")
+            .field("voxel_opacity_mask", &"...")
             .finish()
     }
 }
 
 impl EvaluatedBlock {
+    /// Computes the derived values of a voxel block.
+    fn from_voxels(
+        attributes: BlockAttributes,
+        resolution: Resolution,
+        voxels: GridArray<Evoxel>,
+    ) -> EvaluatedBlock {
+        // Compute color sum from voxels
+        // TODO: Give GridArray an iter() or something
+        // TODO: The color sum actually needs to be weighted by alpha. (Too bad we're not using premultiplied alpha.)
+        // TODO: Should not be counting interior voxels for the color, only visible surfaces.
+        let mut color_sum: Vector4<f32> = Vector4::zero();
+        for position in voxels.grid().interior_iter() {
+            color_sum += voxels[position].color.into();
+        }
+
+        let full_block_grid = Grid::for_block(resolution);
+        EvaluatedBlock {
+            attributes,
+            // The single color is the mean of the actual block colors.
+            color: Rgba::try_from(
+                (color_sum.truncate() / (voxels.grid().volume() as f32))
+                    .extend(color_sum.w / (full_block_grid.volume() as f32)),
+            )
+            .expect("Recursive block color computation produced NaN"),
+            resolution,
+            // TODO wrong test: we want to see if the _faces_ are all opaque but allow hollows
+            opaque: voxels.grid() == full_block_grid
+                && voxels.grid().interior_iter().all(
+                    #[inline(always)]
+                    |p| voxels[p].color.fully_opaque(),
+                ),
+            visible: voxels.grid().interior_iter().any(
+                #[inline(always)]
+                |p| !voxels[p].color.fully_transparent(),
+            ),
+            voxel_opacity_mask: Some(GridArray::from_fn(voxels.grid(), |p| {
+                voxels[p].color.opacity_category()
+            })),
+
+            voxels: Some(voxels),
+        }
+    }
+
     /// Returns whether [`Self::visible`] is true (the block has some visible color/voxels)
     /// or [`BlockAttributes::animation_hint`] indicates that the block might _become_
     /// visible (by change of evaluation result rather than by being replaced).
@@ -715,9 +573,20 @@ impl Evoxel {
         collision: BlockCollision::None,
     };
 
+    /// Construct an [`Evoxel`] which represents the given evaluated block.
+    ///
+    /// This is the same operation as is used for each block/voxel in a [`Block::Recur`].
+    pub fn from_block(block: &EvaluatedBlock) -> Self {
+        Self {
+            color: block.color,
+            selectable: block.attributes.selectable,
+            collision: block.attributes.collision,
+        }
+    }
+
     /// Construct the [`Evoxel`] that would have resulted from evaluating a voxel block
     /// with the given color and default attributes.
-    pub const fn new(color: Rgba) -> Self {
+    pub const fn from_color(color: Rgba) -> Self {
         // Use the values from BlockAttributes's default for consistency.
         // Force constant promotion so that this doesn't look like a
         // feature(const_precise_live_drops) requirement
@@ -776,165 +645,6 @@ impl BlockChange {
     }
 }
 
-/// Contains a [`Block`] and can be stored in a [`Universe`](crate::universe::Universe).
-/// Together with [`Block::Indirect`], this allows mutation of a block definition such
-/// that all its usages follow.
-///
-/// It is a distinct type from [`Block`] in order to ensure that change notifications
-/// will be delivered on any mutation.
-///
-/// To perform a mutation, use [`BlockDefTransaction`].
-#[derive(Debug)]
-pub struct BlockDef {
-    block: Block,
-    // TODO: It might be a good idea to cache EvaluatedBlock here, since we're doing
-    // mutation tracking anyway.
-    notifier: Arc<Notifier<BlockChange>>,
-    block_listen_gate: Gate,
-}
-
-impl BlockDef {
-    pub fn new(block: Block) -> Self {
-        let notifier = Arc::new(Notifier::new());
-        let (gate, block_listener) = Notifier::forwarder(Arc::downgrade(&notifier)).gate();
-        // TODO: Consider making it an error if listening fails. BlockDefTransaction::check will need to follow.
-        let _ = block.listen(block_listener);
-        BlockDef {
-            block,
-            notifier,
-            block_listen_gate: gate,
-        }
-    }
-
-    /// Registers a listener for mutations of any data sources which may affect the
-    /// [`Block::evaluate`] result from blocks defined using this block definition.
-    pub fn listen(
-        &self,
-        listener: impl Listener<BlockChange> + Send + Sync + 'static,
-    ) -> Result<(), RefError> {
-        // TODO: Need to arrange listening to the contained block, and either translate
-        // that here or have our own notifier generate forwardings.
-        self.notifier.listen(listener);
-        Ok(())
-    }
-}
-
-impl Deref for BlockDef {
-    type Target = Block;
-
-    fn deref(&self) -> &Block {
-        &self.block
-    }
-}
-impl AsRef<Block> for BlockDef {
-    fn as_ref(&self) -> &Block {
-        &self.block
-    }
-}
-
-impl Transactional for BlockDef {
-    type Transaction = BlockDefTransaction;
-}
-
-#[derive(Clone, Debug, Default, Eq, Hash, PartialEq)]
-pub struct BlockDefTransaction {
-    // TODO: This struct is the second occurrence (the first is space::CubeTransaction) of a "assign to a mutable location" transaction. If we figure out how to have conveniently _composable_ transactions then we should have an `impl Transaction<&mut T> for Assign<T>` transaction (targeting `&mut` to discourage use otherwise).
-    /// If `None`, no precondition.
-    old: Option<Block>,
-    /// If `None`, no change is made and this transaction is only a precondition.
-    new: Option<Block>,
-}
-
-impl BlockDefTransaction {
-    pub fn expect(old: Block) -> Self {
-        Self {
-            old: Some(old),
-            new: None,
-        }
-    }
-
-    pub fn overwrite(new: Block) -> Self {
-        Self {
-            old: None,
-            new: Some(new),
-        }
-    }
-
-    pub fn replace(old: Block, new: Block) -> Self {
-        Self {
-            old: Some(old),
-            new: Some(new),
-        }
-    }
-}
-
-impl Transaction<BlockDef> for BlockDefTransaction {
-    type CommitCheck = ();
-    type Output = ();
-
-    fn check(
-        &self,
-        target: &BlockDef,
-    ) -> Result<Self::CommitCheck, crate::transaction::PreconditionFailed> {
-        if let Some(old) = &self.old {
-            if **target != *old {
-                return Err(PreconditionFailed {
-                    location: "BlockDef",
-                    problem: "existing block not as expected",
-                });
-            }
-        }
-        Ok(())
-    }
-
-    fn commit(
-        &self,
-        target: &mut BlockDef,
-        (): Self::CommitCheck,
-    ) -> Result<Self::Output, Box<dyn std::error::Error>> {
-        if let Some(new) = &self.new {
-            target.block = new.clone();
-
-            // Swap out the forwarding listener to listen to the new block.
-            let (gate, block_listener) =
-                Notifier::forwarder(Arc::downgrade(&target.notifier)).gate();
-            // TODO: Instead of ignoring the error from listen() here, we can fail the transaction by preparing the listener in check().
-            let _ = target.block.listen(block_listener);
-            target.block_listen_gate = gate; // old gate is now dropped
-
-            target.notifier.notify(BlockChange::new());
-        }
-        Ok(())
-    }
-}
-
-impl Merge for BlockDefTransaction {
-    type MergeCheck = ();
-
-    fn check_merge(
-        &self,
-        other: &Self,
-    ) -> Result<Self::MergeCheck, crate::transaction::TransactionConflict> {
-        if matches!((&self.old, &other.old), (Some(a), Some(b)) if a != b) {
-            return Err(TransactionConflict {});
-        }
-        if matches!((&self.new, &other.new), (Some(a), Some(b)) if a != b) {
-            return Err(TransactionConflict {});
-        }
-        Ok(())
-    }
-
-    fn commit_merge(self, other: Self, (): Self::MergeCheck) -> Self
-    where
-        Self: Sized,
-    {
-        Self {
-            old: self.old.or(other.old),
-            new: self.new.or(other.new),
-        }
-    }
-}
-
 /// Construct a set of [`Block::Recur`] that form a miniature of the given `space`.
 /// The returned [`Space`] contains each of the blocks; its coordinates will correspond to
 /// those of the input, scaled down by `resolution`.
