@@ -1,21 +1,30 @@
-// Copyright 2020-2021 Kevin Reid under the terms of the MIT License as detailed
+// Copyright 2020-2022 Kevin Reid under the terms of the MIT License as detailed
 // in the accompanying file README.md or <https://opensource.org/licenses/MIT>.
 
 #![allow(clippy::bool_assert_comparison)]
 
-use cgmath::EuclideanSpace as _;
 use std::borrow::Cow;
 
+use cgmath::EuclideanSpace as _;
+use pretty_assertions::assert_eq;
+
 use crate::block::{
     builder, AnimationHint, Block, BlockAttributes, BlockBuilder, BlockCollision, BlockDef,
-    BlockDefTransaction, EvalBlockError, Evoxel, Resolution, AIR,
+    BlockDefTransaction, EvalBlockError, EvaluatedBlock, Evoxel, Resolution, RotationPlacementRule,
+    AIR, AIR_EVALUATED,
 };
 use crate::content::{make_some_blocks, make_some_voxel_blocks};
 use crate::listen::{NullListener, Sink};
-use crate::math::{GridPoint, GridRotation, GridVector, Rgb, Rgba};
+use crate::math::{GridPoint, GridRotation, GridVector, OpacityCategory, Rgb, Rgba};
+use crate::raycast::Face;
 use crate::space::{Grid, GridArray, Space, SpacePhysics, SpaceTransaction};
 use crate::universe::Universe;
 
+#[test]
+fn evaluate_air_consistent() {
+    assert_eq!(AIR.evaluate().unwrap(), AIR_EVALUATED);
+}
+
 #[test]
 fn evaluate_opaque_atom_and_attributes() {
     let color = Rgba::new(1.0, 2.0, 3.0, 1.0);
@@ -34,6 +43,10 @@ fn evaluate_opaque_atom_and_attributes() {
     assert_eq!(e.resolution, 1);
     assert_eq!(e.opaque, true);
     assert_eq!(e.visible, true);
+    assert_eq!(
+        e.voxel_opacity_mask,
+        GridArray::from_elements(Grid::for_block(1), [OpacityCategory::Opaque])
+    )
 }
 
 #[test]
@@ -45,6 +58,10 @@ fn evaluate_transparent_atom() {
     assert!(e.voxels.is_none());
     assert_eq!(e.opaque, false);
     assert_eq!(e.visible, true);
+    assert_eq!(
+        e.voxel_opacity_mask,
+        GridArray::from_elements(Grid::for_block(1), [OpacityCategory::Partial])
+    )
 }
 
 #[test]
@@ -55,6 +72,7 @@ fn evaluate_invisible_atom() {
     assert!(e.voxels.is_none());
     assert_eq!(e.opaque, false);
     assert_eq!(e.visible, false);
+    assert_eq!(e.voxel_opacity_mask, None)
 }
 
 #[test]
@@ -92,6 +110,13 @@ fn evaluate_voxels_checked_individually() {
     assert_eq!(e.resolution, resolution);
     assert_eq!(e.opaque, true);
     assert_eq!(e.visible, true);
+    assert_eq!(
+        e.voxel_opacity_mask,
+        GridArray::from_elements(
+            Grid::for_block(resolution),
+            vec![OpacityCategory::Opaque; (resolution as usize).pow(3)]
+        )
+    )
 }
 
 #[test]
@@ -199,6 +224,63 @@ fn evaluate_voxels_zero_resolution() {
     assert_eq!(e.visible, false);
 }
 
+// Unlike other tests, this one asserts the entire `EvaluatedBlock` value because
+// a new field is a potential bug.
+#[test]
+fn evaluate_rotated() {
+    let resolution = 2;
+    let block_grid = Grid::for_block(resolution);
+    let rotation = GridRotation::RYXZ;
+    let mut universe = Universe::new();
+    let color_fn = |cube: GridPoint| {
+        Rgba::new(
+            cube.x as f32,
+            cube.y as f32,
+            cube.z as f32,
+            if cube.y == 0 { 1.0 } else { 0.0 },
+        )
+    };
+    let rotated_color_fn = |cube: GridPoint| {
+        color_fn(
+            rotation
+                .to_positive_octant_matrix(resolution.into())
+                .transform_cube(cube),
+        )
+    };
+    let block = Block::builder()
+        .voxels_fn(&mut universe, resolution, |cube| {
+            // Construct a lower half block with all voxels distinct
+            Block::from(color_fn(cube))
+        })
+        .unwrap()
+        .build();
+    let rotated = block.clone().rotate(rotation);
+    assert_eq!(
+        rotated.evaluate().unwrap(),
+        EvaluatedBlock {
+            attributes: BlockAttributes::default(),
+            color: rgba_const!(0.5, 0.5, 0.5, 0.5),
+            voxels: Some(GridArray::from_fn(block_grid, |cube| {
+                Evoxel {
+                    color: rotated_color_fn(cube),
+                    selectable: true,
+                    collision: BlockCollision::Hard,
+                }
+            })),
+            resolution: 2,
+            opaque: false,
+            visible: true,
+            voxel_opacity_mask: Some(GridArray::from_fn(block_grid, |cube| {
+                if cube.x == 0 {
+                    OpacityCategory::Opaque
+                } else {
+                    OpacityCategory::Invisible
+                }
+            })),
+        }
+    );
+}
+
 /// Tests that the `offset` field of `Block::Recur` is respected.
 #[test]
 fn recur_with_offset() {
@@ -259,7 +341,7 @@ fn indirect_equivalence() {
         .build();
     let eval_bare = block.evaluate();
     let block_def_ref = universe.insert_anonymous(BlockDef::new(block));
-    let eval_def = block_def_ref.borrow().block.evaluate();
+    let eval_def = block_def_ref.borrow().evaluate();
     assert_eq!(eval_bare, eval_def);
 }
 
@@ -385,11 +467,13 @@ fn builder_defaults() {
 fn builder_every_field() {
     let color = Rgba::new(0.1, 0.2, 0.3, 0.4);
     let light_emission = Rgb::new(0.1, 3.0, 0.1);
+    let rotation_rule = RotationPlacementRule::Attach { by: Face::NZ };
     assert_eq!(
         Block::builder()
-            .display_name("hello world")
-            .collision(BlockCollision::None) // TODO: when we have more interesting values, use one
             .color(color)
+            .display_name("hello world")
+            .collision(BlockCollision::Recur)
+            .rotation_rule(rotation_rule)
             .selectable(false)
             .light_emission(light_emission)
             .animation_hint(AnimationHint::TEMPORARY)
@@ -397,7 +481,8 @@ fn builder_every_field() {
         Block::Atom(
             BlockAttributes {
                 display_name: "hello world".into(),
-                collision: BlockCollision::None,
+                collision: BlockCollision::Recur,
+                rotation_rule,
                 selectable: false,
                 light_emission,
                 animation_hint: AnimationHint::TEMPORARY,
@@ -573,6 +658,7 @@ mod txn {
     use super::*;
     use crate::block::BlockDefTransaction;
     use crate::transaction::{Merge, TransactionTester};
+    use pretty_assertions::assert_eq;
 
     #[test]
     fn causes_notification() {
