@@ -1,23 +1,33 @@
-// Copyright 2020-2021 Kevin Reid under the terms of the MIT License as detailed
+// Copyright 2020-2022 Kevin Reid under the terms of the MIT License as detailed
 // in the accompanying file README.md or <https://opensource.org/licenses/MIT>.
 
 //! Projection and view matrices, viewport and aspect ratio, visibility,
 //! raycasting into the scene, etc.
 
 use cgmath::{
-    Deg, EuclideanSpace as _, InnerSpace as _, Matrix as _, Matrix4, Point2, Point3, SquareMatrix,
-    Transform, Vector2, Vector3,
+    Basis3, Decomposed, Deg, EuclideanSpace as _, InnerSpace as _, Matrix4, One as _, Point2,
+    Point3, SquareMatrix, Transform, Vector2, Vector3,
 };
 use itertools::Itertools as _;
 use ordered_float::NotNan;
-use std::convert::TryInto as _;
 
 use crate::math::{Aab, FreeCoordinate, Rgba};
 use crate::raycast::Ray;
 use crate::space::Grid;
 
+mod graphics_options;
+pub use graphics_options::*;
+
 type M = Matrix4<FreeCoordinate>;
 
+/// Representation of a camera viewpoint and orientation, using [`cgmath`] types.
+///
+/// Note that this is treated as a transform **from** the origin looking in the &minus;Z
+/// direction **to** the camera position in the world. This is done so that the
+/// [`Decomposed::disp`] vector is equal to the world position, rather than needing to
+/// be rotated by the view direction.
+pub type ViewTransform = Decomposed<Vector3<FreeCoordinate>, Basis3<FreeCoordinate>>;
+
 /// Defines a viewpoint in/of the world: a viewport (aspect ratio), projection matrix,
 /// and view matrix.
 #[derive(Clone, Debug)]
@@ -25,11 +35,19 @@ pub struct Camera {
     /// Caller-provided options. Always validated by [`GraphicsOptions::repair`].
     options: GraphicsOptions,
 
+    /// Scale factor for scene brightness.
+    pub exposure: NotNan<f32>,
+
     /// Caller-provided viewport.
     viewport: Viewport,
 
-    /// Caller-provided view matrix.
-    view_matrix: M,
+    /// Caller-provided view transform.
+    eye_to_world_transform: ViewTransform,
+
+    /// Inverse of `eye_to_world_transform` as a matrix.
+    /// Might also be called "view matrix".
+    /// Calculated by [`Self::compute_matrices`].
+    world_to_eye_matrix: M,
 
     /// Projection matrix derived from viewport and options.
     /// Calculated by [`Self::compute_matrices`].
@@ -39,9 +57,10 @@ pub struct Camera {
     /// Calculated by [`Self::compute_matrices`].
     view_position: Point3<FreeCoordinate>,
 
-    /// Inverse of `projection * view_matrix`.
+    /// Inverse of `projection * world_to_eye_matrix`.
     /// Calculated by [`Self::compute_matrices`].
     inverse_projection_view: M,
+
     view_frustum_corners: [Point3<FreeCoordinate>; 8],
 }
 
@@ -50,10 +69,12 @@ impl Camera {
     pub fn new(options: GraphicsOptions, viewport: Viewport) -> Self {
         let mut new_self = Self {
             options: options.repair(),
+            exposure: NotNan::one(),
             viewport,
-            view_matrix: M::identity(),
+            eye_to_world_transform: ViewTransform::one(),
 
             // Overwritten immediately by compute_matrices
+            world_to_eye_matrix: M::identity(),
             projection: M::identity(),
             view_position: Point3::origin(),
             inverse_projection_view: M::identity(),
@@ -99,17 +120,26 @@ impl Camera {
         self.options.view_distance.into_inner()
     }
 
-    /// Sets the view matrix.
+    /// Sets the view transform.
     ///
-    /// This matrix is used to determine world coordinates for purposes of
-    /// [`view_position`](Self::view_position) and,
+    /// Besides controlling rendering, this is used to determine world coordinates for purposes
+    /// of [`view_position`](Self::view_position) and
     /// [`project_ndc_into_world`](Self::project_ndc_into_world).
-    /// to determine what world coordinates are.
-    pub fn set_view_matrix(&mut self, view_matrix: M) {
-        if view_matrix != self.view_matrix {
-            self.view_matrix = view_matrix;
-            self.compute_matrices();
+    ///
+    /// The scale currently must be 1.
+    #[track_caller]
+    #[allow(clippy::float_cmp)]
+    pub fn set_view_transform(&mut self, eye_to_world_transform: ViewTransform) {
+        if eye_to_world_transform == self.eye_to_world_transform {
+            return;
         }
+        assert_eq!(
+            eye_to_world_transform.scale, 1.0,
+            "eye_to_world_transform.scale must be equal to 1"
+        );
+
+        self.eye_to_world_transform = eye_to_world_transform;
+        self.compute_matrices();
     }
 
     /// Returns a projection matrix suitable for OpenGL use.
@@ -119,17 +149,17 @@ impl Camera {
 
     /// Returns a view matrix suitable for OpenGL use.
     pub fn view_matrix(&self) -> M {
-        self.view_matrix
+        self.world_to_eye_matrix
     }
 
-    /// Returns the eye position in world coordinates, as set by [`Self::set_view_matrix()`].
+    /// Returns the eye position in world coordinates, as set by [`Camera::set_view_transform()`].
     pub fn view_position(&self) -> Point3<FreeCoordinate> {
         self.view_position
     }
 
     /// Converts a screen position in normalized device coordinates (as produced by
     /// [`Viewport::normalize_nominal_point`]) into a ray in world space.
-    /// Uses the view transformation given by [`set_view_matrix`](Self::set_view_matrix).
+    /// Uses the view transformation given by [`set_view_transform`](Self::set_view_transform).
     pub fn project_ndc_into_world(&self, ndc: Point2<FreeCoordinate>) -> Ray {
         let ndc_near = ndc.to_vec().extend(-1.0).extend(1.0);
         let ndc_far = ndc.to_vec().extend(1.0).extend(1.0);
@@ -148,7 +178,7 @@ impl Camera {
     }
 
     /// Determine whether the given `Aab` is visible in this projection+view.
-    pub(crate) fn aab_in_view(&self, aab: Aab) -> bool {
+    pub fn aab_in_view(&self, aab: Aab) -> bool {
         // Check for the AAB being outside the viewport, using the separating axis theorem.
         // First, check if the view frustum's corner points lie outside the AAB. This is
         // the simpler case since it is axis-aligned.
@@ -215,6 +245,15 @@ impl Camera {
             .unwrap()
     }
 
+    /// Apply postprocessing steps determined by this camera to convert a HDR “scene”
+    /// color into a LDR “image” color. Specifically:
+    ///
+    /// 1. Multiply the input by this camera's exposure value.
+    /// 2. Apply the tone mapping operator specified in [`Camera::options()`].
+    pub fn post_process_color(&self, color: Rgba) -> Rgba {
+        color.map_rgb(|rgb| self.options.tone_mapping.apply(rgb * self.exposure))
+    }
+
     fn compute_matrices(&mut self) {
         self.projection = cgmath::perspective(
             self.fov_y(),
@@ -222,15 +261,15 @@ impl Camera {
             /* near: */ 1. / 32., // half a voxel at resolution=16
             /* far: */ self.view_distance(),
         );
-        self.view_position = Point3::from_vec(
-            self.view_matrix
-                .inverse_transform()
-                .expect("view matrix was not invertible")
-                .transpose()
-                .row(3)
-                .truncate(),
-        );
-        self.inverse_projection_view = (self.projection * self.view_matrix)
+
+        self.world_to_eye_matrix = self.eye_to_world_transform
+            .inverse_transform()
+            .unwrap(/* Inverting cannot fail as long as scale is nonzero */)
+            .into();
+
+        self.view_position = Point3::from_vec(self.eye_to_world_transform.disp);
+
+        self.inverse_projection_view = (self.projection * self.world_to_eye_matrix)
             .inverse_transform()
             .expect("projection and view matrix was not invertible");
 
@@ -308,173 +347,6 @@ impl Viewport {
     // invertible transform.
 }
 
-/// User/debug options for rendering (i.e. not affecting gameplay except informationally).
-/// Not all of these options are applicable to all renderers.
-///
-/// TODO: This may not be the best module location. Possibly it should get its own module.
-#[derive(Clone, Debug, Eq, PartialEq, serde::Deserialize, serde::Serialize)]
-#[serde(default)]
-#[non_exhaustive]
-pub struct GraphicsOptions {
-    /// Whether and how to draw fog obscuring the view distance limit.
-    ///
-    /// TODO: Implement fog in raytracer.
-    pub fog: FogOption,
-
-    /// Field of view, in degrees from top to bottom edge of the viewport.
-    pub fov_y: NotNan<FreeCoordinate>,
-
-    /// Distance, in unit cubes, from the camera to the farthest visible point.
-    ///
-    /// TODO: Implement view distance limit (and fog) in raytracer.
-    pub view_distance: NotNan<FreeCoordinate>,
-
-    /// Style in which to draw the lighting of [`Space`](crate::space::Space)s.
-    /// This does not affect the *computation* of lighting.
-    pub lighting_display: LightingOption,
-
-    /// Method/fidelity to use for transparency.
-    pub transparency: TransparencyOption,
-
-    /// Number of space chunks (16³ groups of blocks) to redraw if needed, per frame.
-    ///
-    /// Does not apply to raytracing.
-    pub chunks_per_frame: u16,
-
-    /// Whether to use frustum culling for drawing only in-view chunks and objects.
-    ///
-    /// This option is for debugging and performance testing and should not have any
-    /// visible effects.
-    pub use_frustum_culling: bool,
-
-    /// Draw boxes around chunk borders and some debug info.
-    pub debug_chunk_boxes: bool,
-
-    /// Draw collision boxes for some objects.
-    pub debug_collision_boxes: bool,
-
-    /// Draw the light rays that contribute to the selected block.
-    pub debug_light_rays_at_cursor: bool,
-}
-
-impl GraphicsOptions {
-    /// Constrain fields to valid/practical values.
-    pub fn repair(mut self) -> Self {
-        self.fov_y = self
-            .fov_y
-            .max(NotNan::new(1.0).unwrap())
-            .min(NotNan::new(189.0).unwrap());
-        self.view_distance = self
-            .view_distance
-            .max(NotNan::new(1.0).unwrap())
-            .min(NotNan::new(10000.0).unwrap());
-        self
-    }
-}
-
-impl Default for GraphicsOptions {
-    fn default() -> Self {
-        Self {
-            fog: FogOption::Compromise,
-            fov_y: NotNan::new(90.).unwrap(),
-            view_distance: NotNan::new(200.).unwrap(),
-            lighting_display: LightingOption::Flat,
-            transparency: TransparencyOption::Volumetric,
-            chunks_per_frame: 4,
-            use_frustum_culling: true,
-            debug_chunk_boxes: false,
-            debug_collision_boxes: false,
-            debug_light_rays_at_cursor: false,
-        }
-    }
-}
-
-#[derive(Clone, Debug, Eq, PartialEq, serde::Deserialize, serde::Serialize)]
-#[non_exhaustive]
-pub enum FogOption {
-    /// No fog: objects will maintain their color and disappear raggedly.
-    None,
-    /// Fog starts just before the view distance ends.
-    Abrupt,
-    /// Compromise between `Abrupt` and `Physical` options.
-    Compromise,
-    /// Almost physically realistic fog of constant density.
-    Physical,
-}
-
-#[derive(Clone, Debug, Eq, PartialEq, serde::Deserialize, serde::Serialize)]
-#[non_exhaustive]
-pub enum LightingOption {
-    /// No lighting: objects will be displayed with their original surface color.
-    None,
-    /// Light is taken from the volume immediately above a cube face.
-    /// Edges between cubes are visible.
-    Flat,
-    /// Light varies across surfaces.
-    Smooth,
-}
-
-/// How to render transparent objects; part of a [`GraphicsOptions`].
-///
-/// Note: There is not yet a consistent interpretation of alpha between the `Surface`
-/// and `Volumetric` options; this will probably be changed in the future in favor
-/// of the volumetric interpretation.
-#[derive(Clone, Debug, Eq, PartialEq, serde::Deserialize, serde::Serialize)]
-#[non_exhaustive]
-pub enum TransparencyOption {
-    /// Conventional transparent surfaces.
-    Surface,
-    /// Accounts for the thickness of material passed through; colors' alpha values are
-    /// interpreted as the opacity of a unit thickness of the material.
-    ///
-    /// TODO: Not implemented in the raytracer.
-    /// TODO: Not implemented correctly for recursive blocks.
-    Volumetric,
-    /// Alpha above or below the given threshold value will be rounded to fully opaque
-    /// or fully transparent, respectively.
-    Threshold(NotNan<f32>),
-}
-
-impl TransparencyOption {
-    /// Replace a color's alpha value according to the requested threshold,
-    /// if any.
-    #[inline]
-    pub(crate) fn limit_alpha(&self, color: Rgba) -> Rgba {
-        match *self {
-            Self::Threshold(t) => {
-                if color.alpha() > t {
-                    color.to_rgb().with_alpha_one()
-                } else {
-                    Rgba::TRANSPARENT
-                }
-            }
-            _ => color,
-        }
-    }
-
-    #[inline]
-    pub(crate) fn will_output_alpha(&self) -> bool {
-        !matches!(self, Self::Threshold(_))
-    }
-}
-
-#[cfg(feature = "arbitrary")]
-impl<'a> arbitrary::Arbitrary<'a> for TransparencyOption {
-    fn arbitrary(u: &mut arbitrary::Unstructured<'a>) -> arbitrary::Result<Self> {
-        use crate::math::arbitrary_notnan;
-        Ok(match u.choose(&[0, 1, 2])? {
-            0 => Self::Surface,
-            1 => Self::Volumetric,
-            2 => Self::Threshold(arbitrary_notnan(u)?),
-            _ => unreachable!(),
-        })
-    }
-
-    fn size_hint(depth: usize) -> (usize, Option<usize>) {
-        arbitrary::size_hint::and(u8::size_hint(depth), Option::<f32>::size_hint(depth))
-    }
-}
-
 /// Calculate an “eye position” (camera position) to view the entire given `grid`.
 ///
 /// `direction` points in the direction the camera should be relative to the space.
@@ -511,11 +383,30 @@ mod tests {
 
     #[test]
     fn camera_view_position() {
+        // This test used to be less trivial when the transform was taken as a matrix
         let mut camera = Camera::new(GraphicsOptions::default(), DUMMY_VIEWPORT);
         let pos = Point3::new(1.0, 2.0, 3.0);
-        camera.set_view_matrix(
-            Matrix4::from_angle_x(Deg(45.0)) * Matrix4::from_translation(-pos.to_vec()),
-        );
+        camera.set_view_transform(Decomposed {
+            scale: 1.0,
+            rot: Basis3::one(),
+            disp: pos.to_vec(),
+        });
         assert_eq!(camera.view_position(), pos);
     }
+
+    #[test]
+    fn post_process() {
+        let mut camera = Camera::new(GraphicsOptions::default(), DUMMY_VIEWPORT);
+
+        // A camera with all default options should pass colors unchanged.
+        let color = rgba_const!(0.1, 0.2, 0.3, 0.4);
+        assert_eq!(camera.post_process_color(color), color);
+
+        // Try exposure
+        camera.exposure = notnan!(0.5);
+        assert_eq!(
+            camera.post_process_color(color),
+            color.map_rgb(|rgb| rgb * 0.5)
+        );
+    }
 }
