@@ -1,11 +1,10 @@
-// Copyright 2020-2021 Kevin Reid under the terms of the MIT License as detailed
+// Copyright 2020-2022 Kevin Reid under the terms of the MIT License as detailed
 // in the accompanying file README.md or <https://opensource.org/licenses/MIT>.
 
 //! [`Inventory`] for storing items.
 
 use std::borrow::Cow;
 use std::collections::BTreeMap;
-use std::error::Error;
 use std::num::NonZeroU16;
 use std::sync::Arc;
 
@@ -14,9 +13,9 @@ use crate::character::{Character, CharacterTransaction, Cursor};
 use crate::inv::{Tool, ToolError, ToolInput};
 use crate::linking::BlockProvider;
 use crate::transaction::{
-    Merge, PreconditionFailed, Transaction, TransactionConflict, UniverseTransaction,
+    CommitError, Merge, PreconditionFailed, Transaction, TransactionConflict, UniverseTransaction,
 };
-use crate::universe::URef;
+use crate::universe::{RefVisitor, URef, VisitRefs};
 use crate::vui::Icons;
 
 /// A collection of [`Tool`]s (items).
@@ -134,6 +133,13 @@ impl Inventory {
     }
 }
 
+impl VisitRefs for Inventory {
+    fn visit_refs(&self, visitor: &mut dyn RefVisitor) {
+        let Self { slots } = self;
+        slots.visit_refs(visitor);
+    }
+}
+
 /// The direct child of [`Inventory`]; a container for any number of identical [`Tool`]s.
 #[derive(Clone, Debug, Eq, Hash, PartialEq)]
 #[non_exhaustive]
@@ -153,6 +159,12 @@ impl Slot {
         }
     }
 
+    /// Temporary const version of [`<Slot as From<Tool>>::from`].
+    #[doc(hidden)]
+    pub const fn one(tool: Tool) -> Self {
+        Self::Stack(Self::COUNT_ONE, tool)
+    }
+
     pub fn icon<'a>(&'a self, predefined: &'a BlockProvider<Icons>) -> Cow<'a, Block> {
         match self {
             Slot::Empty => Cow::Borrowed(&predefined[Icons::EmptySlot]),
@@ -247,6 +259,15 @@ impl From<Option<Tool>> for Slot {
     }
 }
 
+impl VisitRefs for Slot {
+    fn visit_refs(&self, visitor: &mut dyn RefVisitor) {
+        match self {
+            Slot::Empty => {}
+            Slot::Stack(_count, tool) => tool.visit_refs(visitor),
+        }
+    }
+}
+
 /// Specifies a limit on the number of a particular item that should be combined in a
 /// single [`Slot`].
 ///
@@ -383,7 +404,7 @@ impl Transaction<Inventory> for InventoryTransaction {
         &self,
         inventory: &mut Inventory,
         check: Self::CommitCheck,
-    ) -> Result<Self::Output, Box<dyn Error>> {
+    ) -> Result<Self::Output, CommitError> {
         if let Some(InventoryCheck { new, change }) = check {
             assert_eq!(new.len(), inventory.slots.len());
             inventory.slots = new;
@@ -586,7 +607,7 @@ mod tests {
         ];
         const MAX: u16 = u16::MAX;
         let gen_slots = move || {
-            IntoIterator::into_iter([
+            [
                 0,
                 1,
                 2,
@@ -598,8 +619,9 @@ mod tests {
                 MAX - 2,
                 MAX - 1,
                 MAX,
-            ])
-            .cartesian_product(IntoIterator::into_iter(tools.clone()))
+            ]
+            .into_iter()
+            .cartesian_product(tools.clone())
             .map(|(count, item)| Slot::stack(count, item))
         };
         for slot1_in in gen_slots() {
