@@ -1,4 +1,4 @@
-// Copyright 2020-2021 Kevin Reid under the terms of the MIT License as detailed
+// Copyright 2020-2022 Kevin Reid under the terms of the MIT License as detailed
 // in the accompanying file README.md or <https://opensource.org/licenses/MIT>.
 
 //! [`Tool`] and related.
@@ -7,14 +7,15 @@ use std::borrow::Cow;
 use std::sync::Arc;
 use std::{fmt, hash};
 
-use crate::block::{Block, AIR};
+use crate::block::{Block, RotationPlacementRule, AIR};
 use crate::character::{Character, CharacterTransaction, Cursor};
 use crate::inv::{InventoryTransaction, StackLimit};
 use crate::linking::BlockProvider;
-use crate::math::GridPoint;
+use crate::math::{GridPoint, GridRotation};
+use crate::raycast::Face;
 use crate::space::SpaceTransaction;
 use crate::transaction::{Merge, Transaction, UniverseTransaction};
-use crate::universe::{RefError, URef};
+use crate::universe::{RefError, RefVisitor, URef, VisitRefs};
 use crate::vui::Icons;
 
 /// A `Tool` is an object which a character can use to have some effect in the game,
@@ -43,6 +44,12 @@ pub enum Tool {
     /// Copy block from space to inventory.
     CopyFromSpace,
 
+    /// Allows flight.
+    ///
+    /// TODO: This should probably be a feature a tool can have rather than a
+    /// single-purpose item, but we don't yet have a plan for programmable items.
+    Jetpack { active: bool },
+
     /// A tool which calls an arbitrary function.
     ExternalAction {
         // TODO: Rework this so that the external component gets to update the icon.
@@ -79,17 +86,12 @@ impl Tool {
     ) -> Result<(Option<Self>, UniverseTransaction), ToolError> {
         match self {
             Self::Activate => {
-                // TODO: We have nothing to activate yet.
-                // (But this error also needs work.)
-                Err(ToolError::NotUsable)
-            }
-            Self::ExternalAction { ref function, .. } => {
-                if let Some(f) = &function.0 {
-                    f(input);
-                    Ok((Some(self), Default::default()))
-                } else {
-                    Err(ToolError::NotUsable) // TODO: communicate permanent error
-                }
+                let cursor = input.cursor()?;
+                // TODO: cursor is probably not _exactly_ the right set of data that should be passed on
+                Ok((
+                    Some(self),
+                    SpaceTransaction::activate_block(cursor.place.cube).bind(cursor.space.clone()),
+                ))
             }
             Self::RemoveBlock { keep } => {
                 let cursor = input.cursor()?;
@@ -112,15 +114,12 @@ impl Tool {
             Self::Block(ref block) => {
                 let cursor = input.cursor()?;
                 let block = block.clone();
-                Ok((None, input.set_cube(cursor.place.adjacent(), AIR, block)?))
+                Ok((None, input.place_block(cursor, AIR, block)?))
             }
             Self::InfiniteBlocks(ref block) => {
                 let cursor = input.cursor()?;
                 let block = block.clone();
-                Ok((
-                    Some(self),
-                    input.set_cube(cursor.place.adjacent(), AIR, block)?,
-                ))
+                Ok((Some(self), input.place_block(cursor, AIR, block)?))
             }
             Self::CopyFromSpace => {
                 let cursor = input.cursor()?;
@@ -130,6 +129,18 @@ impl Tool {
                         .produce_item(Tool::InfiniteBlocks(cursor.block.clone().unspecialize()))?,
                 ))
             }
+            Self::Jetpack { active } => Ok((
+                Some(Self::Jetpack { active: !active }),
+                UniverseTransaction::default(),
+            )),
+            Self::ExternalAction { ref function, .. } => {
+                if let Some(f) = &function.0 {
+                    f(input);
+                    Ok((Some(self), UniverseTransaction::default()))
+                } else {
+                    Err(ToolError::NotUsable) // TODO: communicate permanent error
+                }
+            }
         }
     }
 
@@ -158,13 +169,15 @@ impl Tool {
     pub fn icon<'a>(&'a self, predefined: &'a BlockProvider<Icons>) -> Cow<'a, Block> {
         match self {
             Self::Activate => Cow::Borrowed(&predefined[Icons::Activate]),
-            Self::ExternalAction { icon, .. } => Cow::Borrowed(icon),
             // TODO: Give Remove different icons
             Self::RemoveBlock { keep: _ } => Cow::Borrowed(&predefined[Icons::Delete]),
             // TODO: Once blocks have behaviors, we need to defuse them for this use.
             // TODO: InfiniteBlocks should have a different name and appearance
             Self::Block(block) | Self::InfiniteBlocks(block) => Cow::Borrowed(block),
             Self::CopyFromSpace => Cow::Borrowed(&predefined[Icons::CopyFromSpace]),
+            Self::Jetpack { active: true } => Cow::Borrowed(&predefined[Icons::JetpackOn]),
+            Self::Jetpack { active: false } => Cow::Borrowed(&predefined[Icons::JetpackOff]),
+            Self::ExternalAction { icon, .. } => Cow::Borrowed(icon),
         }
     }
 
@@ -178,11 +191,28 @@ impl Tool {
             Tool::Block(_) => Standard,
             Tool::InfiniteBlocks(_) => One,
             Tool::CopyFromSpace => One,
+            Tool::Jetpack { .. } => One,
             Tool::ExternalAction { .. } => One,
         }
     }
 }
 
+impl VisitRefs for Tool {
+    fn visit_refs(&self, visitor: &mut dyn RefVisitor) {
+        match self {
+            Tool::Activate => {}
+            Tool::RemoveBlock { .. } => {}
+            Tool::Block(block) => block.visit_refs(visitor),
+            Tool::InfiniteBlocks(block) => block.visit_refs(visitor),
+            Tool::CopyFromSpace => {}
+            Tool::Jetpack { active: _ } => {}
+            Tool::ExternalAction { function: _, icon } => {
+                icon.visit_refs(visitor);
+            }
+        }
+    }
+}
+
 /// Resources available to a `Tool` to perform its function.
 ///
 /// This is intended to provide future extensibility compared to having a complex
@@ -218,6 +248,40 @@ impl ToolInput {
         )
     }
 
+    /// As [`Self::set_cube`] but also applying rotation (or other transformations
+    /// in the future) specified by the block's attributes
+    fn place_block(
+        &self,
+        cursor: &Cursor,
+        old_block: Block,
+        new_block: Block,
+    ) -> Result<UniverseTransaction, ToolError> {
+        let rotation = match new_block
+            .evaluate()
+            .map_err(|e| ToolError::Internal(e.to_string()))? // TODO: better error typing here
+            .attributes
+            .rotation_rule
+        {
+            RotationPlacementRule::Never => GridRotation::IDENTITY,
+            RotationPlacementRule::Attach { by: attached_face } => {
+                // TODO: RotationPlacementRule should control the "up" axis choices
+                GridRotation::from_to(attached_face, cursor.place.face.opposite(), Face::PY)
+                    .or_else(|| {
+                        GridRotation::from_to(attached_face, cursor.place.face.opposite(), Face::PX)
+                    })
+                    .or_else(|| {
+                        GridRotation::from_to(attached_face, cursor.place.face.opposite(), Face::PZ)
+                    })
+                    .unwrap_or(GridRotation::IDENTITY)
+            }
+        };
+        self.set_cube(
+            cursor.place.adjacent(),
+            old_block,
+            new_block.rotate(rotation),
+        )
+    }
+
     /// Returns a [`Cursor`] indicating what blocks the tool should act on, if it is
     /// a sort of tool that acts on blocks. If there is no [`Cursor`], because of aim
     /// or because of being used in a context where there cannot be any aiming, returns
@@ -276,7 +340,7 @@ pub enum ToolError {
 /// TODO: relocate this type once we figure out where it belongs.
 /// TODO: Probably they should be their own kind of UniverseMember, so that they can
 /// be reattached in the future.
-pub struct EphemeralOpaque<T: ?Sized>(Option<Arc<T>>);
+pub struct EphemeralOpaque<T: ?Sized>(pub(crate) Option<Arc<T>>);
 
 impl<T: ?Sized> From<Arc<T>> for EphemeralOpaque<T> {
     fn from(contents: Arc<T>) -> Self {
@@ -331,17 +395,17 @@ impl<'a, T: arbitrary::Arbitrary<'a>> arbitrary::Arbitrary<'a> for EphemeralOpaq
 
 #[cfg(test)]
 mod tests {
-    use std::error::Error;
-
     use super::*;
     use crate::character::cursor_raycast;
-    use crate::content::make_some_blocks;
+    use crate::content::{make_some_blocks, make_some_voxel_blocks};
     use crate::inv::Slot;
-    use crate::math::FreeCoordinate;
+    use crate::math::{FreeCoordinate, GridRotation};
     use crate::raycast::Ray;
     use crate::raytracer::print_space;
     use crate::space::Space;
     use crate::universe::{UBorrow, URef, Universe};
+    use pretty_assertions::assert_eq;
+    use std::error::Error;
 
     #[derive(Debug)]
     struct ToolTester {
@@ -406,7 +470,10 @@ mod tests {
 
         /// As `equip_and_use_tool`, but also commit the transaction.
         /// TODO: Needs a better error return type (requires Transaction to do so).
-        fn equip_use_commit(&mut self, stack: impl Into<Slot>) -> Result<(), Box<dyn Error>> {
+        fn equip_use_commit(
+            &mut self,
+            stack: impl Into<Slot>,
+        ) -> Result<(), Box<dyn Error + Send + Sync>> {
             let transaction = self.equip_and_use_tool(stack)?;
             transaction.execute(&mut self.universe)?;
             Ok(())
@@ -439,20 +506,19 @@ mod tests {
     }
 
     #[test]
-    fn use_activate_noop() {
+    fn use_activate() {
         let [existing] = make_some_blocks();
         let tester = ToolTester::new(|space| {
             space.set((1, 0, 0), &existing).unwrap();
         });
         assert_eq!(
             tester.equip_and_use_tool(Tool::Activate),
-            Err(ToolError::NotUsable)
+            Ok(SpaceTransaction::activate_block(GridPoint::new(1, 0, 0))
+                .bind(tester.space_ref.clone()))
         );
-    }
 
-    #[test]
-    fn use_activate_effect() {
-        // TODO: When Tool::Activate actually gains some effects, test them here.
+        // Tool::Activate currently has no cases where it fails
+        // (unless the transaction fails), so there are no tests for that.
     }
 
     #[test]
@@ -559,6 +625,40 @@ mod tests {
         }
     }
 
+    /// TODO: Expand this test to exhaustively test all rotation placement rules?
+    #[test]
+    fn use_block_automatic_rotation() {
+        let [existing] = make_some_blocks();
+        let mut tester = ToolTester::new(|space| {
+            space.set((1, 0, 0), &existing).unwrap();
+        });
+
+        // Make a block with a rotation rule
+        let [mut tool_block] = make_some_voxel_blocks(&mut tester.universe);
+        if let Block::Recur {
+            ref mut attributes, ..
+        } = tool_block
+        {
+            attributes.rotation_rule = RotationPlacementRule::Attach { by: Face::NZ };
+        } else {
+            unreachable!();
+        }
+
+        // TODO: For more thorough testing, we need to be able to control ToolTester's choice of ray
+        let transaction = tester
+            .equip_and_use_tool(Tool::InfiniteBlocks(tool_block.clone()))
+            .unwrap();
+        assert_eq!(
+            transaction,
+            SpaceTransaction::set_cube(
+                [0, 0, 0],
+                Some(AIR),
+                Some(tool_block.clone().rotate(GridRotation::CLOCKWISE))
+            )
+            .bind(tester.space_ref.clone())
+        );
+    }
+
     /// Note: This is more of a test of [`Inventory`] and [`Slot`] stack management
     /// than the tool.
     #[test]
