@@ -1,13 +1,13 @@
-// Copyright 2020-2021 Kevin Reid under the terms of the MIT License as detailed
+// Copyright 2020-2022 Kevin Reid under the terms of the MIT License as detailed
 // in the accompanying file README.md or <https://opensource.org/licenses/MIT>.
 
 //! [`Face`] type and related items.
 //! This module is private but reexported by its parent.
 
+use std::ops::{Index, IndexMut};
+
 use cgmath::{BaseNum, Transform, Vector3};
 pub use ordered_float::{FloatIsNan, NotNan};
-use std::convert::TryFrom;
-use std::ops::{Index, IndexMut};
 
 use crate::math::*;
 
@@ -19,6 +19,7 @@ use crate::math::*;
 #[allow(clippy::upper_case_acronyms)]
 #[allow(clippy::exhaustive_enums)]
 #[derive(Clone, Copy, Debug, Hash, Eq, PartialEq)]
+#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
 #[repr(u8)]
 pub enum Face {
     /// The interior volume of a cube, or an undefined direction. Corresponds to the vector `(0, 0, 0)`.
@@ -199,6 +200,34 @@ impl Face {
         }
     }
 
+    /// Dot product of this face as a unit vector and the given vector,
+    /// implemented by selecting the relevant component.
+    ///
+    /// ```
+    /// use cgmath::{Vector3, InnerSpace};
+    /// use all_is_cubes::math::Face;
+    ///
+    /// let sample_vector = Vector3::new(1.0, 2.0, 5.0_f64);
+    /// for face in Face::ALL_SEVEN {
+    ///     assert_eq!(face.dot(sample_vector), face.normal_vector().dot(sample_vector));
+    /// }
+    /// ```
+    #[inline]
+    pub fn dot<S>(self, vector: Vector3<S>) -> S
+    where
+        S: Zero + std::ops::Neg<Output = S>,
+    {
+        match self {
+            Face::Within => S::zero(),
+            Face::NX => -vector.x,
+            Face::NY => -vector.y,
+            Face::NZ => -vector.z,
+            Face::PX => vector.x,
+            Face::PY => vector.y,
+            Face::PZ => vector.z,
+        }
+    }
+
     /// Returns a homogeneous transformation matrix which, if given points on the square
     /// with x ∈ [0, scale], y ∈ [0, scale] and z = 0, converts them to points that lie
     /// on the faces of the cube with x ∈ [0, scale], y ∈ [0, scale], and z ∈ [0, scale].
@@ -267,7 +296,6 @@ impl TryFrom<GridVector> for Face {
     ///
     /// ```
     /// use all_is_cubes::math::{Face, GridVector};
-    /// use std::convert::TryFrom;
     ///
     /// // A Face may be converted from its normal vector.
     /// for face in Face::ALL_SEVEN {
@@ -350,6 +378,22 @@ impl<V> FaceMap<V> {
         }
     }
 
+    /// Returns a vector containing the values for each negative face.
+    pub fn negatives(self) -> Vector3<V>
+    where
+        V: Copy,
+    {
+        Vector3::new(self.nx, self.ny, self.nz)
+    }
+
+    /// Returns a vector containing the values for each positive face.
+    pub fn positives(self) -> Vector3<V>
+    where
+        V: Copy,
+    {
+        Vector3::new(self.px, self.py, self.pz)
+    }
+
     /// Iterate over the map entries by reference.
     pub fn iter<'s>(&'s self) -> impl Iterator<Item = (Face, &V)> + 's {
         Face::ALL_SEVEN.iter().copied().map(move |f| (f, &self[f]))
@@ -368,7 +412,8 @@ impl<V> FaceMap<V> {
     }
 
     pub fn into_values_iter(self) -> impl Iterator<Item = V> {
-        std::array::IntoIter::new(self.into_values())
+        // TODO: eliminate this as not really useful in Rust 2021
+        self.into_values().into_iter()
     }
 
     /// Transform values.
@@ -510,7 +555,7 @@ impl Geometry for CubeFace {
         ];
         // TODO: this is a messy kludge and really we should be stealing corner points
         // from the AAB instead, but there isn't yet a good way to do that.
-        output.extend(IntoIterator::into_iter(X_POINTS).map(|p| {
+        output.extend(X_POINTS.into_iter().map(|p| {
             (
                 (face_matrix.transform_point(p))
                     .map(|c| (FreeCoordinate::from(c) - 0.5) * (1. + expansion * 2.) + 0.5)
