@@ -1,4 +1,4 @@
-// Copyright 2020-2021 Kevin Reid under the terms of the MIT License as detailed
+// Copyright 2020-2022 Kevin Reid under the terms of the MIT License as detailed
 // in the accompanying file README.md or <https://opensource.org/licenses/MIT>.
 
 use cgmath::{EuclideanSpace as _, InnerSpace as _, Point3, Vector3, Zero};
@@ -8,13 +8,14 @@ use std::fmt;
 use super::collision::{
     aab_raycast, collide_along_ray, find_colliding_cubes, nudge_on_ray, Contact,
 };
-use crate::apps::Tick;
 use crate::block::BlockCollision;
 use crate::math::{Aab, Face, FreeCoordinate, Geometry as _};
+use crate::physics::{StopAt, POSITION_EPSILON};
 use crate::raycast::Ray;
 use crate::space::Space;
+use crate::time::Tick;
 use crate::transaction::{
-    Merge, PreconditionFailed, Transaction, TransactionConflict, Transactional,
+    CommitError, Merge, PreconditionFailed, Transaction, TransactionConflict, Transactional,
 };
 use crate::util::{ConciseDebug, CustomFormat, StatusText};
 
@@ -235,7 +236,7 @@ impl Body {
             movement_ignoring_collision,
             self.collision_box,
             collision_callback,
-            true,
+            StopAt::NotAlreadyColliding,
         );
 
         if let Some(collision) = collision {
@@ -325,41 +326,59 @@ impl Body {
         space: &Space,
         direction: Vector3<FreeCoordinate>,
     ) -> Option<(Point3<FreeCoordinate>, NotNan<FreeCoordinate>)> {
-        let ray = Ray::new(self.position, direction);
-        // TODO: upper bound on distance to try
-        'raycast: for ray_step in aab_raycast(self.collision_box, ray, true) {
-            let adjusted_segment = nudge_on_ray(
-                self.collision_box,
-                ray.scale_direction(ray_step.t_distance()),
-                ray_step.face(),
-                1,
-                true,
-            );
-            let step_aab = self
-                .collision_box
-                .translate(adjusted_segment.unit_endpoint().to_vec());
-            for cube in step_aab.round_up_to_grid().interior_iter() {
-                // TODO: refactor to combine this with other collision attribute tests
-                match space.get_evaluated(cube).attributes.collision {
-                    BlockCollision::Hard => {
-                        // Not a clear space
-                        continue 'raycast;
-                    }
-                    BlockCollision::None => {}
-                    BlockCollision::Recur => {
-                        // TODO: Either check collision, or continue
-                        //continue 'raycast;
+        if false {
+            // TODO: This attempted implementation does not work, causing lots of falling into
+            // blocks. But if we can fix the bugs, it will make push-out actually work with
+            // recursive blocks. (It's why I added StopAt::EmptySpace.)
+
+            let direction = direction.normalize(); // TODO: set this to a max distance
+            let ray = Ray::new(self.position, direction);
+
+            let end =
+                collide_along_ray(space, ray, self.collision_box, |_| {}, StopAt::EmptySpace)?;
+
+            let nudged_distance = end.t_distance + POSITION_EPSILON;
+            Some((
+                ray.scale_direction(nudged_distance).unit_endpoint(),
+                NotNan::new(nudged_distance).ok()?,
+            ))
+        } else {
+            let ray = Ray::new(self.position, direction);
+            // TODO: upper bound on distance to try
+            'raycast: for ray_step in aab_raycast(self.collision_box, ray, true) {
+                let adjusted_segment = nudge_on_ray(
+                    self.collision_box,
+                    ray.scale_direction(ray_step.t_distance()),
+                    ray_step.face(),
+                    1,
+                    true,
+                );
+                let step_aab = self
+                    .collision_box
+                    .translate(adjusted_segment.unit_endpoint().to_vec());
+                for cube in step_aab.round_up_to_grid().interior_iter() {
+                    // TODO: refactor to combine this with other collision attribute tests
+                    match space.get_evaluated(cube).attributes.collision {
+                        BlockCollision::Hard => {
+                            // Not a clear space
+                            continue 'raycast;
+                        }
+                        BlockCollision::None => {}
+                        BlockCollision::Recur => {
+                            // TODO: Either check collision, or continue
+                            //continue 'raycast;
+                        }
                     }
                 }
+                // No collisions, so we can use this.
+                return Some((
+                    adjusted_segment.unit_endpoint(),
+                    NotNan::new(ray_step.t_distance() * direction.magnitude()).ok()?,
+                ));
             }
-            // No collisions, so we can use this.
-            return Some((
-                adjusted_segment.unit_endpoint(),
-                NotNan::new(ray_step.t_distance() * direction.magnitude()).ok()?,
-            ));
-        }
 
-        None
+            None
+        }
     }
 
     /// Returns the body's collision box in world coordinates
@@ -492,11 +511,7 @@ impl Transaction<Body> for BodyTransaction {
         Ok(())
     }
 
-    fn commit(
-        &self,
-        body: &mut Body,
-        _: Self::CommitCheck,
-    ) -> Result<(), Box<dyn std::error::Error>> {
+    fn commit(&self, body: &mut Body, _: Self::CommitCheck) -> Result<(), CommitError> {
         body.yaw += self.delta_yaw;
         Ok(())
     }
