@@ -1,4 +1,4 @@
-// Copyright 2020-2021 Kevin Reid under the terms of the MIT License as detailed
+// Copyright 2020-2022 Kevin Reid under the terms of the MIT License as detailed
 // in the accompanying file README.md or <https://opensource.org/licenses/MIT>.
 
 //! Algorithm for raycasting through voxel grids.
@@ -207,16 +207,32 @@ impl Raycaster {
     /// This makes the iterator finite: [`next()`](Self::next) will return [`None`]
     /// forevermore once there are no more cubes intersecting the grid to report.
     pub fn within_grid(mut self, grid: Grid) -> Self {
+        self.set_bound(grid);
+        self
+    }
+
+    /// Like `within_grid` but not moving self.
+    ///
+    /// TODO: This function was added for the needs of the raytracer. Think about API design more.
+    pub(crate) fn set_bound(&mut self, grid: Grid) {
         if self.grid == None {
             self.grid = Some(grid);
         } else {
             unimplemented!("multiple uses of .within_grid()");
         }
-        if false {
-            // Not actually faster, so disabled for now. See function doc.
-            self.fast_forward();
-        }
-        self
+        self.fast_forward();
+    }
+
+    /// Cancels a previous [`Raycaster::within_grid`], allowing the raycast to proceed
+    /// an arbitrary distance.
+    ///
+    /// Note: The effect of calling `within_grid()` and then `remove_bound()` without an
+    /// intervening `next()` is not currently guaranteed.
+    ///
+    /// TODO: This function was added for the needs of the raytracer. Think about API design more.
+
+    pub(crate) fn remove_bound(&mut self) {
+        self.grid = None;
     }
 
     #[inline(always)]
@@ -314,12 +330,7 @@ impl Raycaster {
     }
 
     /// In the case where the current position is outside the grid but might intersect
-    /// the grid later, attempt to move the position to intersect sooner, in a fashion
-    /// we'd hope is more efficient than single stepping.
-    ///
-    /// However, benchmarks have not shown it to be of substantial benefit, so it's
-    /// not enabled currently. Future use cases might differ, so I'm keeping the code
-    /// around and compiling at least "until 1.0"...
+    /// the grid later, attempt to move the position to intersect sooner.
     #[inline(always)]
     fn fast_forward(&mut self) {
         let grid: Grid = self.grid.unwrap();
