@@ -1,4 +1,4 @@
-// Copyright 2020-2021 Kevin Reid under the terms of the MIT License as detailed
+// Copyright 2020-2022 Kevin Reid under the terms of the MIT License as detailed
 // in the accompanying file README.md or <https://opensource.org/licenses/MIT>.
 
 //! Raytracer for [`Space`]s.
@@ -7,115 +7,187 @@
 //!
 //! The original reason this exists is that I thought “we have [`raycast`](crate::raycast),
 //! and that's nearly all the work, so why not?” Secondarily, it was written before
-//! the [mesh-based renderer](crate::lum), and was useful as a cross-check since
+//! [the mesh-based renderer](crate::mesh), and was useful as a cross-check since
 //! it is much simpler. It continues to serve as a “reference implementation” and is used
 //! by the terminal UI and in unit tests via [`print_space`].
 
-use std::convert::TryFrom as _;
 use std::fmt;
 
-use cgmath::{EuclideanSpace as _, InnerSpace as _, Point2, Vector3};
+use cgmath::{EuclideanSpace as _, InnerSpace as _, Point2, Vector2, Vector3};
 use cgmath::{Point3, Vector4};
-use ouroboros::self_referencing;
+use ordered_float::NotNan;
 #[cfg(feature = "rayon")]
 use rayon::iter::{IntoParallelIterator as _, ParallelIterator as _};
 
-use crate::block::{Evoxel, Resolution};
-use crate::camera::{Camera, GraphicsOptions};
+use crate::block::{Evoxel, Resolution, AIR};
+use crate::camera::{Camera, GraphicsOptions, TransparencyOption};
 use crate::math::{smoothstep, GridCoordinate};
 use crate::math::{Face, FreeCoordinate, GridPoint, Rgb, Rgba};
 use crate::raycast::Ray;
-use crate::space::{GridArray, PackedLight, Space};
+use crate::space::{BlockIndex, GridArray, PackedLight, Space, SpaceBlockData};
 
 mod pixel_buf;
 pub use pixel_buf::*;
 
 mod surface;
-use surface::{Surface, SurfaceIter, TraceStep};
+use surface::{DepthIter, DepthStep, Span, Surface, SurfaceIter, TraceStep};
 // TODO: pub use surface::*;
 
 mod text;
 pub use text::*;
 
-/// Precomputed data for raytracing a single frame of a single Space, and bearer of the
-/// methods for actually performing raytracing.
-pub struct SpaceRaytracer<P: PixelBuf>(SpaceRaytracerImpl<P>);
+/// Precomputed data for raytracing a single frame of a single [`Space`], and bearer of
+/// the methods for actually performing raytracing.
+pub struct SpaceRaytracer<D: RtBlockData> {
+    blocks: Vec<TracingBlock<D>>,
+    cubes: GridArray<TracingCubeData>,
 
-/// Helper struct for [`SpaceRaytracer`] so the details of [`ouroboros::self_referencing`]
-/// aren't exposed.
-#[self_referencing]
-struct SpaceRaytracerImpl<P: PixelBuf> {
-    blocks: Box<[TracingBlock<P::BlockData>]>,
-    #[borrows(blocks)]
-    #[covariant]
-    cubes: GridArray<TracingCubeData<'this, P::BlockData>>,
-
-    options: GraphicsOptions,
+    graphics_options: GraphicsOptions,
+    custom_options: D::Options,
     sky_color: Rgb,
+    sky_data: D,
     packed_sky_color: PackedLight,
 }
 
-impl<P: PixelBuf> SpaceRaytracer<P> {
+impl<D: RtBlockData> SpaceRaytracer<D> {
     /// Snapshots the given [`Space`] to prepare for raytracing it.
-    pub fn new(space: &Space, options: GraphicsOptions) -> Self {
+    pub fn new(
+        space: &Space,
+        graphics_options: GraphicsOptions,
+        custom_options: D::Options,
+    ) -> Self {
+        let options = RtOptionsRef {
+            graphics_options: &graphics_options,
+            custom_options: &custom_options,
+        };
         let sky_color = space.physics().sky_color;
-        #[allow(clippy::borrowed_box)]
-        SpaceRaytracer(
-            SpaceRaytracerImplBuilder {
-                blocks: prepare_blocks::<P>(space),
-                cubes_builder: |blocks: &Box<[TracingBlock<P::BlockData>]>| {
-                    prepare_cubes::<P>(blocks, space)
-                },
-                options,
-                sky_color,
-                packed_sky_color: sky_color.into(),
-            }
-            .build(),
-        )
+        SpaceRaytracer {
+            blocks: space
+                .block_data()
+                .iter()
+                .map(|sbd| TracingBlock::<D>::from_block(options, sbd))
+                .collect(),
+            cubes: prepare_cubes(space),
+            sky_color,
+            sky_data: D::sky(options),
+            packed_sky_color: sky_color.into(),
+
+            graphics_options,
+            custom_options,
+        }
     }
 
     /// Computes a single image pixel from the given ray.
-    pub fn trace_ray(&self, ray: Ray) -> (P::Pixel, RaytraceInfo) {
-        self.0.with(|impl_fields| {
-            let mut s: TracingState<P> = TracingState::default();
-            for step in SurfaceIter::new(self, ray) {
-                if s.count_step_should_stop() {
-                    break;
-                }
+    pub fn trace_ray<P: PixelBuf<BlockData = D>>(&self, ray: Ray) -> (P, RaytraceInfo) {
+        let options = RtOptionsRef {
+            graphics_options: &self.graphics_options,
+            custom_options: &self.custom_options,
+        };
 
-                use TraceStep::*;
-                match step {
-                    Invisible | EnterBlock => {
-                        // Side effect: called count_step_should_stop.
+        let t_to_absolute_distance = ray.direction.magnitude();
+        let mut s: TracingState<P> = TracingState::default();
+        let surface_iter = SurfaceIter::new(self, ray);
+
+        // Use the more expensive volumetric tracing strategy only if we need it.
+        match self.graphics_options.transparency {
+            TransparencyOption::Volumetric => {
+                for step in DepthIter::new(surface_iter) {
+                    if s.count_step_should_stop(options) {
+                        break;
                     }
-                    EnterSurface(surface) => {
-                        debug_assert!(!surface.diffuse_color.fully_transparent());
-                        // TODO: To implement TransparencyOption::Volumetric we need to peek forward to the next change of color and find the distance between them, but only if the alpha is not 0 or 1.
-                        s.trace_through_surface(surface, self);
+
+                    match step {
+                        DepthStep::Invisible => {
+                            // Side effect: called count_step_should_stop.
+                        }
+                        DepthStep::Span(Span {
+                            mut surface,
+                            exit_t_distance,
+                        }) => {
+                            debug_assert!(!surface.diffuse_color.fully_transparent());
+
+                            let thickness = ((exit_t_distance - surface.t_distance)
+                                * t_to_absolute_distance)
+                                as f32;
+
+                            // Convert alpha to transmittance (light transmitted / light received).
+                            let unit_transmittance =
+                                1.0 - surface.diffuse_color.alpha().into_inner();
+                            // Adjust transmittance for the thickness relative to an assumed 1.0 thickness.
+                            let depth_transmittance = unit_transmittance.powf(thickness);
+                            // Convert back to alpha.
+                            // TODO: skip NaN check ... this may require refactoring Surface usage.
+                            // We might also benefit from an "UncheckedRgba" concept.
+                            surface.diffuse_color = surface
+                                .diffuse_color
+                                .to_rgb()
+                                .with_alpha(NotNan::new(1.0 - depth_transmittance).unwrap());
+
+                            s.trace_through_surface(surface, self);
+                        }
                     }
                 }
             }
-            s.finish(*impl_fields.sky_color)
-        })
+            _ => {
+                for step in surface_iter {
+                    if s.count_step_should_stop(options) {
+                        break;
+                    }
+
+                    use TraceStep::*;
+                    match step {
+                        Invisible { .. } | EnterBlock { .. } => {
+                            // Side effect: called count_step_should_stop.
+                        }
+                        EnterSurface(surface) => {
+                            debug_assert!(!surface.diffuse_color.fully_transparent());
+                            s.trace_through_surface(surface, self);
+                        }
+                    }
+                }
+            }
+        }
+        s.finish(self.sky_color, &self.sky_data)
     }
 
     /// Compute a full image.
     ///
-    /// The returned `[P::Pixel]` is in the usual left-right then top-bottom raster order;
+    /// The returned data is in the usual left-right then top-bottom raster order;
     /// its dimensions are `camera.framebuffer_size`.
     ///
+    /// `encoder` may be used to transform the output of the `PixelBuf`.
+    ///
     /// TODO: Add a mechanism for incrementally rendering into a mutable buffer instead of
     /// all-at-once into a newly allocated one, for interactive use.
-    pub fn trace_scene_to_image(&self, camera: &Camera) -> (Box<[P::Pixel]>, RaytraceInfo) {
+    pub fn trace_scene_to_image<P, E, O>(
+        &self,
+        camera: &Camera,
+        encoder: E,
+    ) -> (Box<[O]>, RaytraceInfo)
+    where
+        P: PixelBuf<BlockData = D>,
+        E: Fn(P) -> O + Send + Sync,
+        O: Send + Sync,
+    {
         // This wrapper function ensures that the two implementations have consistent
         // signatures.
-        self.trace_scene_to_image_impl(camera)
+        self.trace_scene_to_image_impl(camera, encoder)
     }
 
     #[cfg(feature = "rayon")]
-    fn trace_scene_to_image_impl(&self, camera: &Camera) -> (Box<[P::Pixel]>, RaytraceInfo) {
+    fn trace_scene_to_image_impl<P, E, O>(
+        &self,
+        camera: &Camera,
+        encoder: E,
+    ) -> (Box<[O]>, RaytraceInfo)
+    where
+        P: PixelBuf<BlockData = D>,
+        E: Fn(P) -> O + Send + Sync,
+        O: Send + Sync,
+    {
         let viewport = camera.viewport();
         let viewport_size = viewport.framebuffer_size.map(|s| s as usize);
+        let encoder = &encoder; // make shareable
 
         let output_iterator = (0..viewport_size.y)
             .into_par_iter()
@@ -123,19 +195,30 @@ impl<P: PixelBuf> SpaceRaytracer<P> {
                 let y = viewport.normalize_fb_y(ych);
                 (0..viewport_size.x).into_par_iter().map(move |xch| {
                     let x = viewport.normalize_fb_x(xch);
-                    self.trace_ray(camera.project_ndc_into_world(Point2::new(x, y)))
+                    let (pixel, info) =
+                        self.trace_ray(camera.project_ndc_into_world(Point2::new(x, y)));
+                    (encoder(pixel), info)
                 })
             })
             .flatten();
 
-        let (image, info_sum): (Vec<P::Pixel>, rayon_helper::ParExtSum<RaytraceInfo>) =
+        let (image, info_sum): (Vec<O>, rayon_helper::ParExtSum<RaytraceInfo>) =
             output_iterator.unzip();
 
         (image.into_boxed_slice(), info_sum.result())
     }
 
     #[cfg(not(feature = "rayon"))]
-    fn trace_scene_to_image_impl(&self, camera: &Camera) -> (Box<[P::Pixel]>, RaytraceInfo) {
+    fn trace_scene_to_image_impl<P, E, O>(
+        &self,
+        camera: &Camera,
+        encoder: E,
+    ) -> (Box<[O]>, RaytraceInfo)
+    where
+        P: PixelBuf<BlockData = D>,
+        E: Fn(P) -> O + Send + Sync,
+        O: Send + Sync,
+    {
         let viewport = camera.viewport();
         let viewport_size = viewport.framebuffer_size.map(|s| s as usize);
         let mut image = Vec::with_capacity(viewport.pixel_count().expect("image too large"));
@@ -148,7 +231,7 @@ impl<P: PixelBuf> SpaceRaytracer<P> {
                 let (pixel, info) =
                     self.trace_ray(camera.project_ndc_into_world(Point2::new(x, y)));
                 total_info += info;
-                image.push(pixel);
+                image.push(encoder(pixel));
             }
         }
 
@@ -157,28 +240,22 @@ impl<P: PixelBuf> SpaceRaytracer<P> {
 
     #[inline]
     fn get_packed_light(&self, cube: GridPoint) -> PackedLight {
-        self.0.with(|impl_fields| {
-            impl_fields
-                .cubes
-                .get(cube)
-                .map(|b| b.lighting)
-                .unwrap_or(*impl_fields.packed_sky_color)
-        })
+        self.cubes
+            .get(cube)
+            .map(|b| b.lighting)
+            .unwrap_or(self.packed_sky_color)
     }
 
     #[inline]
     fn get_lighting(&self, cube: GridPoint) -> Rgb {
-        self.0.with(|impl_fields| {
-            impl_fields
-                .cubes
-                .get(cube)
-                .map(|b| b.lighting.value())
-                .unwrap_or(*impl_fields.sky_color)
-        })
+        self.cubes
+            .get(cube)
+            .map(|b| b.lighting.value())
+            .unwrap_or(self.sky_color)
     }
 
     fn get_interpolated_light(&self, point: Point3<FreeCoordinate>, face: Face) -> Rgb {
-        // This implementation is duplicated in GLSL at src/lum/shaders/fragment.glsl
+        // This implementation is duplicated in GLSL at all-is-cubes-gpu/src/shaders/fragment.glsl
 
         // About half the size of the smallest permissible voxel.
         let above_surface_epsilon = 0.5 / 256.0;
@@ -255,36 +332,39 @@ impl<P: PixelBuf> SpaceRaytracer<P> {
     }
 }
 
-impl<P: PixelBuf<Pixel = String>> SpaceRaytracer<P> {
-    /// Raytrace to text, using any [`PixelBuf`] whose output is [`String`].
+/// Text-specific methods.
+impl<D: RtBlockData> SpaceRaytracer<D> {
+    /// Raytrace to text, using any [`PixelBuf`] whose output can be [`String`].
     ///
     /// `F` is the function accepting the output, and `E` is the type of error it may
     /// produce. This function-based interface is intended to abstract over the
     /// inconvenient difference between [`std::io::Write`] and [`std::fmt::Write`].
     ///
     /// After each line (row) of the image, `write(line_ending)` will be called.
-    pub fn trace_scene_to_text<F, E>(
+    pub fn trace_scene_to_text<P, F, E>(
         &self,
         camera: &Camera,
         line_ending: &str,
         write: F,
     ) -> Result<RaytraceInfo, E>
     where
+        P: PixelBuf<BlockData = D> + Into<String>,
         F: FnMut(&str) -> Result<(), E>,
     {
         // This wrapper function ensures that the two implementations have consistent
         // signatures.
-        self.trace_scene_to_text_impl(camera, line_ending, write)
+        self.trace_scene_to_text_impl::<P, F, E>(camera, line_ending, write)
     }
 
     #[cfg(feature = "rayon")]
-    fn trace_scene_to_text_impl<F, E>(
+    fn trace_scene_to_text_impl<P, F, E>(
         &self,
         camera: &Camera,
         line_ending: &str,
         mut write: F,
     ) -> Result<RaytraceInfo, E>
     where
+        P: PixelBuf<BlockData = D> + Into<String>,
         F: FnMut(&str) -> Result<(), E>,
     {
         let viewport = camera.viewport();
@@ -297,7 +377,9 @@ impl<P: PixelBuf<Pixel = String>> SpaceRaytracer<P> {
                     .into_par_iter()
                     .map(move |xch| {
                         let x = viewport.normalize_fb_x(xch);
-                        self.trace_ray(camera.project_ndc_into_world(Point2::new(x, y)))
+                        let (buf, info) =
+                            self.trace_ray::<P>(camera.project_ndc_into_world(Point2::new(x, y)));
+                        (buf.into(), info)
                     })
                     .chain(Some((line_ending.to_owned(), RaytraceInfo::default())).into_par_iter())
             })
@@ -305,19 +387,20 @@ impl<P: PixelBuf<Pixel = String>> SpaceRaytracer<P> {
 
         let (text, info_sum): (String, rayon_helper::ParExtSum<RaytraceInfo>) =
             output_iterator.unzip();
-        write(text.as_ref())?;
+        write(text.as_str())?;
 
         Ok(info_sum.result())
     }
 
     #[cfg(not(feature = "rayon"))]
-    fn trace_scene_to_text_impl<F, E>(
+    fn trace_scene_to_text_impl<P, F, E>(
         &self,
         camera: &Camera,
         line_ending: &str,
         mut write: F,
     ) -> Result<RaytraceInfo, E>
     where
+        P: PixelBuf<BlockData = D> + Into<String>,
         F: FnMut(&str) -> Result<(), E>,
     {
         let mut total_info = RaytraceInfo::default();
@@ -328,27 +411,46 @@ impl<P: PixelBuf<Pixel = String>> SpaceRaytracer<P> {
             let y = viewport.normalize_fb_y(ych);
             for xch in 0..viewport_size.x {
                 let x = viewport.normalize_fb_x(xch);
-                let (text, info) = self.trace_ray(camera.project_ndc_into_world(Point2::new(x, y)));
+                let (buf, info) =
+                    self.trace_ray::<P>(camera.project_ndc_into_world(Point2::new(x, y)));
                 total_info += info;
-                write(text.as_ref())?;
+                write(buf.into().as_str())?;
             }
             write(line_ending)?;
         }
 
         Ok(total_info)
     }
+
+    pub fn trace_scene_to_string<P>(&self, camera: &Camera, line_ending: &str) -> String
+    where
+        P: PixelBuf<BlockData = D> + Into<String>,
+    {
+        let mut out = String::with_capacity(
+            camera.viewport().framebuffer_size.dot(Vector2::new(1, 1)) as usize,
+        );
+        self.trace_scene_to_text::<P, _, _>(camera, line_ending, |s| {
+            out.push_str(s);
+            Ok::<(), std::convert::Infallible>(())
+        })
+        .unwrap();
+        out
+    }
 }
 
-impl<P: PixelBuf> fmt::Debug for SpaceRaytracer<P> {
+impl<D: RtBlockData> fmt::Debug for SpaceRaytracer<D>
+where
+    D: fmt::Debug,
+    D::Options: fmt::Debug,
+{
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        self.0.with(|impl_fields| {
-            f.debug_struct("SpaceRaytracer")
-                .field("blocks.len", &impl_fields.blocks.len())
-                .field("cubes.grid", &impl_fields.cubes.grid())
-                .field("options", &impl_fields.options)
-                .field("sky_color", &impl_fields.sky_color)
-                .finish_non_exhaustive()
-        })
+        f.debug_struct("SpaceRaytracer")
+            .field("blocks.len", &self.blocks.len())
+            .field("cubes.grid", &self.cubes.grid())
+            .field("graphics_options", &self.graphics_options)
+            .field("custom_options", &self.custom_options)
+            .field("sky_color", &self.sky_color)
+            .finish_non_exhaustive()
     }
 }
 
@@ -379,51 +481,51 @@ impl std::iter::Sum for RaytraceInfo {
     }
 }
 
-/// Get block data out of [`Space`] (which is not [`Sync`], and not specialized for our
-/// efficient use).
+/// Get cube data out of [`Space`].
 #[inline]
-fn prepare_blocks<P: PixelBuf>(space: &Space) -> Box<[TracingBlock<P::BlockData>]> {
-    space
-        .block_data()
-        .iter()
-        .map(|block_data| {
-            let evaluated = block_data.evaluated();
-            let pixel_block_data = P::compute_block_data(block_data);
-            if let Some(ref voxels) = evaluated.voxels {
-                TracingBlock::Recur(pixel_block_data, evaluated.resolution, voxels.clone())
-            } else {
-                TracingBlock::Atom(pixel_block_data, evaluated.color)
-            }
-        })
-        .collect()
-}
-
-/// Get cube data out of [`Space`] (which is not [`Sync`], and not specialized for our
-/// efficient use).
-#[inline]
-#[allow(clippy::ptr_arg)] // no benefit
-fn prepare_cubes<'a, P: PixelBuf>(
-    indexed_block_data: &'a [TracingBlock<P::BlockData>],
-    space: &Space,
-) -> GridArray<TracingCubeData<'a, P::BlockData>> {
-    space.extract(space.grid(), |index, _block, lighting| TracingCubeData {
-        block: &indexed_block_data[index.unwrap() as usize],
-        lighting,
+fn prepare_cubes(space: &Space) -> GridArray<TracingCubeData> {
+    space.extract(space.grid(), |index, block_data, lighting| {
+        TracingCubeData {
+            block_index: index.unwrap(),
+            lighting,
+            always_invisible: block_data.block() == &AIR,
+        }
     })
 }
 
 #[derive(Clone, Debug)]
-struct TracingCubeData<'a, B: 'static> {
-    block: &'a TracingBlock<B>,
+struct TracingCubeData {
+    block_index: BlockIndex,
     lighting: PackedLight,
+    /// True if the block is [`AIR`].
+    ///
+    /// This special information allows us to skip an indirect memory access in this
+    /// extremely common case. We could generalize it to any block which is fully
+    /// invisible, but only if *the block is not an indirection* since if it is, the
+    /// block data could change without signaling a cube change, and currently we don't
+    /// have a mechanism to obtain that information from the Space.
+    always_invisible: bool,
 }
 
 #[derive(Clone, Debug)]
-enum TracingBlock<B: 'static> {
-    Atom(B, Rgba),
-    Recur(B, Resolution, GridArray<Evoxel>),
+enum TracingBlock<D> {
+    Atom(D, Rgba),
+    Recur(D, Resolution, GridArray<Evoxel>),
 }
 
+impl<D: RtBlockData> TracingBlock<D> {
+    fn from_block(options: RtOptionsRef<'_, D::Options>, block_data: &SpaceBlockData) -> Self {
+        let evaluated = block_data.evaluated();
+        let pixel_block_data = D::from_block(options, block_data);
+        if let Some(ref voxels) = evaluated.voxels {
+            TracingBlock::Recur(pixel_block_data, evaluated.resolution, voxels.clone())
+        } else {
+            TracingBlock::Atom(pixel_block_data, evaluated.color)
+        }
+    }
+}
+
+/// Holds a [`PixelBuf`] and other per-ray state.
 #[derive(Clone, Debug, Default)]
 struct TracingState<P: PixelBuf> {
     /// Number of cubes traced through -- controlled by the caller, so not necessarily
@@ -433,28 +535,30 @@ struct TracingState<P: PixelBuf> {
 }
 impl<P: PixelBuf> TracingState<P> {
     #[inline]
-    fn count_step_should_stop(&mut self) -> bool {
+    fn count_step_should_stop(
+        &mut self,
+        options: RtOptionsRef<'_, <P::BlockData as RtBlockData>::Options>,
+    ) -> bool {
         self.cubes_traced += 1;
         if self.cubes_traced > 1000 {
             // Abort excessively long traces.
             self.pixel_buf = Default::default();
             self.pixel_buf
-                .add(Rgba::new(1.0, 1.0, 1.0, 1.0), &P::error_block_data());
+                .add(Rgba::new(1.0, 1.0, 1.0, 1.0), &P::BlockData::error(options));
             true
         } else {
             self.pixel_buf.opaque()
         }
     }
 
-    fn finish(mut self, sky_color: Rgb) -> (P::Pixel, RaytraceInfo) {
+    fn finish(mut self, sky_color: Rgb, sky_data: &P::BlockData) -> (P, RaytraceInfo) {
         if self.cubes_traced == 0 {
             // Didn't intersect the world at all. Draw these as plain background.
             // TODO: Switch to using the sky color, unless debugging options are set.
             self.pixel_buf.hit_nothing();
         }
 
-        self.pixel_buf
-            .add(sky_color.with_alpha_one(), &P::sky_block_data());
+        self.pixel_buf.add(sky_color.with_alpha_one(), sky_data);
 
         // Debug visualization of number of raytracing steps.
         // TODO: Make this togglable and less of a kludge — we'd like to be able to mix with
@@ -463,12 +567,12 @@ impl<P: PixelBuf> TracingState<P> {
             self.pixel_buf = Default::default();
             self.pixel_buf.add(
                 (rgb_const!(0.02, 0.002, 0.0) * self.cubes_traced as f32).with_alpha_one(),
-                &P::sky_block_data(),
+                sky_data,
             );
         }
 
         (
-            self.pixel_buf.result(),
+            self.pixel_buf,
             RaytraceInfo {
                 cubes_traced: self.cubes_traced,
             },
@@ -480,7 +584,7 @@ impl<P: PixelBuf> TracingState<P> {
     fn trace_through_surface(
         &mut self,
         surface: Surface<'_, P::BlockData>,
-        rt: &SpaceRaytracer<P>,
+        rt: &SpaceRaytracer<P::BlockData>,
     ) {
         if let Some(color) = surface.to_lit_color(rt) {
             self.pixel_buf.add(color, surface.block_data);
@@ -488,6 +592,9 @@ impl<P: PixelBuf> TracingState<P> {
     }
 }
 
+pub use updating::*;
+mod updating;
+
 #[cfg(feature = "rayon")]
 mod rayon_helper {
     use rayon::iter::{IntoParallelIterator, ParallelExtend, ParallelIterator as _};
