@@ -1,4 +1,4 @@
-// Copyright 2020-2021 Kevin Reid under the terms of the MIT License as detailed
+// Copyright 2020-2022 Kevin Reid under the terms of the MIT License as detailed
 // in the accompanying file README.md or <https://opensource.org/licenses/MIT>.
 
 //! That which contains many blocks.
@@ -10,7 +10,6 @@ use std::sync::{Arc, Mutex, Weak};
 
 use cgmath::Vector3;
 
-use crate::apps::Tick;
 use crate::behavior::{Behavior, BehaviorSet};
 use crate::block::*;
 use crate::character::Spawn;
@@ -18,9 +17,9 @@ use crate::content::palette;
 use crate::drawing::DrawingPlane;
 use crate::listen::{Gate, Listener, Notifier};
 use crate::math::*;
-use crate::space::lighting::opaque_for_light_computation;
+use crate::time::Tick;
 use crate::transaction::{Transaction as _, UniverseTransaction};
-use crate::universe::URef;
+use crate::universe::{RefVisitor, URef, VisitRefs};
 use crate::util::ConciseDebug;
 use crate::util::{CustomFormat, StatusText};
 
@@ -30,12 +29,8 @@ pub use builder::SpaceBuilder;
 mod grid;
 pub use grid::*;
 
-mod lighting;
-pub use lighting::LightUpdatesInfo;
-
-mod light_data;
-pub use light_data::PackedLight;
-use light_data::{LightUpdateQueue, PackedLightScalar};
+mod light;
+pub use light::*;
 
 mod space_txn;
 pub use space_txn::*;
@@ -69,7 +64,8 @@ pub struct Space {
     light_update_queue: LightUpdateQueue,
     /// Debug log of the updated cubes from last frame.
     /// Empty unless this debug function is enabled.
-    pub(crate) last_light_updates: Vec<GridPoint>,
+    #[doc(hidden)] // pub to be used by all-is-cubes-gpu
+    pub last_light_updates: Vec<GridPoint>,
 
     /// Global characteristics such as the behavior of light and gravity.
     physics: SpacePhysics,
@@ -661,8 +657,8 @@ impl Space {
         &self.spawn
     }
 
-    pub fn spawn_mut(&mut self) -> &mut Spawn {
-        &mut self.spawn
+    pub fn set_spawn(&mut self, spawn: Spawn) {
+        self.spawn = spawn;
     }
 
     pub fn add_behavior<B>(&mut self, behavior: B)
@@ -809,6 +805,31 @@ impl<T: Into<GridPoint>> std::ops::Index<T> for Space {
     }
 }
 
+impl VisitRefs for Space {
+    fn visit_refs(&self, visitor: &mut dyn RefVisitor) {
+        let Space {
+            grid: _,
+            block_to_index: _,
+            block_data,
+            contents: _,
+            lighting: _,
+            light_update_queue: _,
+            last_light_updates: _,
+            physics: _,
+            packed_sky_color: _,
+            behaviors,
+            spawn,
+            notifier: _,
+            todo: _,
+        } = self;
+        for SpaceBlockData { block, .. } in block_data {
+            block.visit_refs(visitor);
+        }
+        behaviors.visit_refs(visitor);
+        spawn.visit_refs(visitor);
+    }
+}
+
 impl SpaceBlockData {
     /// A `SpaceBlockData` value used to represent out-of-bounds or placeholder
     /// situations. The block is [`AIR`] and the count is always zero.
@@ -937,11 +958,7 @@ impl Default for SpacePhysics {
 impl<'a> arbitrary::Arbitrary<'a> for SpacePhysics {
     fn arbitrary(u: &mut arbitrary::Unstructured<'a>) -> arbitrary::Result<Self> {
         Ok(Self {
-            gravity: Vector3::new(
-                arbitrary_notnan(u)?,
-                arbitrary_notnan(u)?,
-                arbitrary_notnan(u)?,
-            ),
+            gravity: Vector3::new(u.arbitrary()?, u.arbitrary()?, u.arbitrary()?),
             sky_color: u.arbitrary()?,
             light: u.arbitrary()?,
         })
@@ -957,6 +974,7 @@ impl<'a> arbitrary::Arbitrary<'a> for SpacePhysics {
         ])
     }
 }
+
 /// Method used to compute the illumination of individual blocks in a [`Space`].
 #[non_exhaustive]
 #[derive(Clone, Debug, Eq, Hash, PartialEq)]
@@ -1011,7 +1029,7 @@ pub enum SetCubeError {
 
 /// Description of a change to a [`Space`] for use in listeners.
 #[derive(Clone, Debug, Eq, Hash, PartialEq)]
-#[non_exhaustive]
+#[allow(clippy::exhaustive_enums)] // any change will probably be breaking anyway
 pub enum SpaceChange {
     // TODO: This set of names is not very clear and self-consistent.
     /// The block at the given location was replaced.
