@@ -1,10 +1,9 @@
-// Copyright 2020-2021 Kevin Reid under the terms of the MIT License as detailed
+// Copyright 2020-2022 Kevin Reid under the terms of the MIT License as detailed
 // in the accompanying file README.md or <https://opensource.org/licenses/MIT>.
 
 //! Axis-aligned integer-coordinate box volumes ([`Grid`]), arrays bounded by them
 //! ([`GridArray`]), and related.
 
-use std::convert::TryFrom;
 use std::fmt;
 use std::iter::FusedIterator;
 use std::ops::Range;
@@ -120,9 +119,13 @@ impl Grid {
     ///
     /// If you need such a grid at a position other than the origin, use
     /// [`Grid::translate`].
-    pub fn for_block(resolution: Resolution) -> Grid {
-        let size = GridCoordinate::from(resolution);
-        Grid::new([0, 0, 0], [size, size, size])
+    pub const fn for_block(resolution: Resolution) -> Grid {
+        // The limited range of Resolution lets us write this as const.
+        let size = resolution as GridCoordinate;
+        Grid {
+            lower_bounds: GridPoint::new(0, 0, 0),
+            sizes: GridVector::new(size, size, size),
+        }
     }
 
     /// Generate a `Grid` whose volume is as specified or smaller.
@@ -131,8 +134,6 @@ impl Grid {
         u: &mut arbitrary::Unstructured<'_>,
         volume: usize,
     ) -> arbitrary::Result<Self> {
-        use std::convert::TryInto;
-
         // Pick sizes within the volume constraint.
         let mut limit: GridCoordinate = volume.try_into().unwrap_or(GridCoordinate::MAX);
         let size_1 = u.int_in_range(0..=limit)?;
@@ -567,6 +568,7 @@ impl Grid {
     /// );
     /// ```
     #[inline]
+    #[track_caller] // TODO: better error reporting
     pub fn expand(self, deltas: FaceMap<GridCoordinate>) -> Self {
         use Face::*;
         let l = self.lower_bounds();
